# 10/24 CSC 401 Examples

# Load necessary libraries
library(datasets)

# Load the Iris dataset
data(iris)

# Prepare the data by removing the Species column to separate features and labels
iris_data <- iris[, -5]
iris_labels <- iris[, 5]

# Euclidean distance function
euclidean_distance <- function(x, y) {
  return(sqrt(sum((x - y)^2)))
}

# KNN classification function
knn_classify <- function(train_data, train_labels, test_data, k = 3) {
  predictions <- c()
  
  for (i in 1:nrow(test_data)) {
    # Calculate distances from the test point to all training points
    distances <- apply(train_data, 1, function(row) euclidean_distance(row, test_data[i, ]))
    
    # Sort the distances and get the indices of the k nearest neighbors
    k_nearest_indices <- order(distances)[1:k]
    
    # Get the labels of the k nearest neighbors
    k_nearest_labels <- train_labels[k_nearest_indices]
    
    # Determine the most common label (mode) among the k nearest neighbors
    predicted_label <- names(sort(table(k_nearest_labels), decreasing = TRUE))[1]
    
    # Append the prediction to the list of predictions
    predictions <- c(predictions, predicted_label)
  }
  
  return(predictions)
}

# Split the data into training and testing sets (80/20 split)
set.seed(123) # for reproducibility
train_indices <- sample(1:nrow(iris_data), size = 0.8 * nrow(iris_data))

train_data <- iris_data[train_indices, ]
train_labels <- iris_labels[train_indices]

test_data <- iris_data[-train_indices, ]
test_labels <- iris_labels[-train_indices]

# Run the KNN classifier with k=3
k <- 3
predictions <- knn_classify(train_data, train_labels, test_data, k = k)

# Calculate accuracy
accuracy <- sum(predictions == test_labels) / length(test_labels)
print(paste("Accuracy:", accuracy))

# Confusion matrix
table(Predicted = predictions, Actual = test_labels)

#regression
# Load the mtcars dataset
data(mtcars)

# Prepare the data by selecting features and the target variable
mtcars_data <- mtcars[, c("disp", "hp", "wt")]
mtcars_labels <- mtcars[, "mpg"]

# Euclidean distance function (same as above)
euclidean_distance <- function(x, y) {
  return(sqrt(sum((x - y)^2)))
}

# KNN regression function
knn_regress <- function(train_data, train_labels, test_data, k = 3) {
  predictions <- c()
  
  for (i in 1:nrow(test_data)) {
    # Calculate distances from the test point to all training points
    distances <- apply(train_data, 1, function(row) euclidean_distance(row, test_data[i, ]))
    
    # Sort the distances and get the indices of the k nearest neighbors
    k_nearest_indices <- order(distances)[1:k]
    
    # Get the labels (mpg values) of the k nearest neighbors
    k_nearest_labels <- train_labels[k_nearest_indices]
    
    # Predict the mean of the k nearest labels
    predicted_value <- mean(k_nearest_labels)
    
    # Append the prediction to the list of predictions
    predictions <- c(predictions, predicted_value)
  }
  
  return(predictions)
}

# Split the data into training and testing sets (80/20 split)
set.seed(123) # for reproducibility
train_indices <- sample(1:nrow(mtcars_data), size = 0.8 * nrow(mtcars_data))

train_data <- mtcars_data[train_indices, ]
train_labels <- mtcars_labels[train_indices]

test_data <- mtcars_data[-train_indices, ]
test_labels <- mtcars_labels[-train_indices]

# Run the KNN regressor with k=3
k <- 3
predictions <- knn_regress(train_data, train_labels, test_data, k = k)

# Calculate Mean Absolute Error (MAE)
mae <- mean(abs(predictions - test_labels))
print(paste("Mean Absolute Error (MAE):", mae))

# Plot predictions vs actual values
plot(test_labels, predictions, xlab = "Actual MPG", ylab = "Predicted MPG", main = "KNN Regression")
abline(0, 1, col = "red") # Reference line

#comparing linear regression to knn
# Step 1: Fit a Linear Regression Model
linear_model <- lm(mpg ~ disp + hp + wt, data = mtcars)

# Step 2: Generate Predictions from the Linear Regression Model
linear_predictions <- predict(linear_model, newdata = test_data)

# Step 3: Generate Predictions from the KNN Regression Model (using k=3 as before)
k <- 3
knn_predictions <- knn_regress(train_data, train_labels, test_data, k = k)

# Step 4: Create a Comparison Plot
plot(test_labels, linear_predictions, col = "blue", pch = 16,
     xlab = "Actual MPG", ylab = "Predicted MPG",
     main = "Comparison of Linear Regression and KNN Predictions")
points(test_labels, knn_predictions, col = "red", pch = 16)

# Add a legend
legend("topleft", legend = c("Linear Regression", "KNN Regression"),
       col = c("blue", "red"), pch = 16)

# Add a reference line for perfect prediction
abline(0, 1, col = "black", lty = 2)

