# 10/3 CSC 401 Examples

matern <- function(r, nu, theta) 
{
  rat <- r*sqrt(2*nu/theta)
  C <- (2^(1 - nu))/gamma(nu) * rat^nu * besselK(rat, nu)
  C[is.nan(C)] <- 1
  return(C)
}

r <- seq(eps, 3, length=100)
plot(r, matern(r, nu=1/2, theta=1), type="l", ylab="k(r,nu)")
lines(r, matern(r, nu=2, theta=1), lty=2, col=2)
lines(r, matern(r, nu=10, theta=1), lty=3, col=3)
legend("topright", c("nu=1/2", "n=2", "nu=10"), lty=1:3, col=1:3, bty="n")

X <- seq(0, 10, length=100)
R <- sqrt(distance(X))
K0.5 <- matern(R, nu=1/2, theta=1)
K2 <- matern(R, nu=2, theta=1) 
K10 <- matern(R, nu=10, theta=1) + diag(eps, 100)

par(mfrow=c(1,3))
matplot(X, t(rmvnorm(3,sigma=K0.5)), type="l", col=1, lty=1, 
        xlab="x", ylab="y, nu=1/2")
matplot(X, t(rmvnorm(3,sigma=K2)), type="l", col=2, lty=2, 
        xlab="x", ylab="y, nu=2")
matplot(X, t(rmvnorm(3,sigma=K10)), type="l", col=3, lty=3, xlab="x", 
        ylab="y, nu=10")
par(mfrow=c(1,1))

F <- rmvnorm(1, sigma=K2)
dF <- (F[-1] - F[-length(F)])/(X[2] - X[1])
d2F <- (dF[-1] - dF[-length(dF)])/(X[2] - X[1])

plot(X, F, type="l", lwd=2, xlim=c(0,13), ylim=c(-5,5), 
     ylab="F and derivatives")
lines(X[-1], dF, col=2, lty=2)
lines(X[-(1:2)], d2F, col=3, lty=3)
legend("topright", c("F", "dF", "d2F"), lty=1:3, col=1:3, bty="n")

#power exponential
powerexp <- function(r, alpha, theta) 
{
  C <- exp(-(r/sqrt(theta))^alpha)
  C[is.nan(C)] <- 1
  return(C)
}

plot(r, powerexp(r, alpha=1.5, theta=1), type="l", ylab="k(r,alpha)")
lines(r, powerexp(r, alpha=1.9, theta=1), lty=2, col=2)
lines(r, powerexp(r, alpha=2, theta=1), lty=3, col=3)
legend("topright", c("alpha=1.5", "alpha=1.9", "alpha=2"), 
       lty=1:3, col=1:3, bty="n")

Ka1.5 <- powerexp(R, alpha=1.5, theta=1)
Ka1.9 <- powerexp(R, alpha=1.9, theta=1)
Ka2 <- powerexp(R, alpha=2, theta=1) + diag(eps, nrow(R))
par(mfrow=c(1,3))
ylab <- paste0("y, alpha=", c(1.5, 1.9, 2))
matplot(X, t(rmvnorm(3, sigma=Ka1.5)), type="l", col=1, lty=1, 
        xlab="x", ylab=ylab[1])
matplot(X, t(rmvnorm(3, sigma=Ka1.9)), type="l", col=2, lty=2, 
        xlab="x", ylab=ylab[2])
matplot(X, t(rmvnorm(3, sigma=Ka2)), type="l", col=3, lty=3, 
        xlab="x", ylab=ylab[3])
par(mfrow=c(1,1))

#rational quadratic
ratquad <- function(r, alpha, theta) 
{
  C <- (1 + r^2/(2 * alpha * theta))^(-alpha)
  C[is.nan(C)] <- 1
  return(C)
}

par(mfrow=c(1,2))
plot(r, ratquad(r, alpha=1/2, theta=1), type="l", 
     ylab="k(r,alpha)", ylim=c(0,1))
lines(r, ratquad(r, alpha=2, theta=1), lty=2, col=2)
lines(r, ratquad(r, alpha=10, theta=1), lty=3, col=3)
legend("topright", c("alpha=1/2", "alpha=2", "alpha=10"), lty=1:3, 
       col=1:3, bty="n")
Eps <- diag(eps, nrow(R))
plot(X, rmvnorm(1, sigma=ratquad(R, alpha=1/2, theta=1) + Eps), type="l", 
     col=1, lty=1, xlab="x", ylab="y", ylim=c(-2.5,2.5))
lines(X, rmvnorm(1, sigma=ratquad(R, alpha=2, theta=1) + Eps), col=2, lty=2)
lines(X, rmvnorm(1, sigma=ratquad(R, alpha=10, theta=1) + Eps), type="l", 
      col=3, lty=3)
par(mfrow=c(1,1))

#noise
x <- seq(0, 10, length=40)
ytrue <- (sin(pi*x/5) + 0.2*cos(4*pi*x/5))
y <- ytrue + rnorm(length(ytrue), sd=0.2)

g <- seq(0.001, 0.4, length=100)
theta <- seq(0.1, 4, length=100)
grid <- expand.grid(theta, g)

library(laGP)
ll <- rep(NA, nrow(grid))
xx <- seq(0, 10, length=100)
pm <- matrix(NA, nrow=nrow(grid), ncol=length(xx))
psd <- matrix(NA, nrow=nrow(grid), ncol=length(xx))
for(i in 1:nrow(grid)) {
  gpi <- newGP(matrix(x, ncol=1), y, d=grid[i,1], g=grid[i,2])
  p <- predGP(gpi, matrix(xx, ncol=1), lite=TRUE)
  pm[i,] <- p$mean
  psd[i,] <- sqrt(p$s2)
  ll[i] <- llikGP(gpi)
  deleteGP(gpi)
}
l <- exp(ll - max(ll))

image(theta, g, matrix(l, ncol=length(theta)), col=cols)
contour(theta, g, matrix(l, ncol=length(g)), add=TRUE)

q1 <- pm + qnorm(0.95, sd=psd)
q2 <- pm + qnorm(0.05, sd=psd)

plot(x,y, ylim=c(range(q1, q2)))
matlines(xx, t(pm), col=rgb(0,0,0,alpha=(l/max(l))/2), lty=1)
matlines(xx, t(q1), col=rgb(1,0,0,alpha=(l/max(l))/2), lty=2)
matlines(xx, t(q2), col=rgb(1,0,0,alpha=(l/max(l))/2), lty=2)

#convolutions
ell <- 10
n <- 100
X <- seq(0, 10, length=n)
omega <- seq(-1, 11, length=ell)
K <- matrix(NA, ncol=ell, nrow=n)
for(j in 1:ell) K[,j] <- dnorm(X, omega[j])

beta <- matrix(rnorm(3*ell), ncol=3)
F <- K %*% beta
matplot(X, F, type="l", lwd=2, lty=1, col="gray", 
        xlim=c(-1,11), ylab="f(x)")
abline(v=omega, lty=2, lwd=0.5)

n <- length(x)
K <- as.data.frame(matrix(NA, ncol=ell, nrow=n))
for(j in 1:ell) K[,j] <- dnorm(x, omega[j]) 
names(K) <- paste0("omega", 1:ell)

fit <- lm(y ~ . -1, data=K)

xx <- seq(-1, 11, length=100)
KK <- as.data.frame(matrix(NA, ncol=ell, nrow=length(xx)))
for(j in 1:ell) KK[,j] <- dnorm(xx, omega[j]) 
names(KK) <- paste0("omega", 1:ell)

p <- predict(fit, newdata=KK, interval="prediction")

plot(x, y)
lines(xx, p[,1])
lines(xx, p[,2], col=2, lty=2)
lines(xx, p[,3], col=2, lty=2)

for(j in 1:ell) K[,j] <- dnorm(x, omega[j], sd=0.5) 
fit <- lm(y ~ . -1, data=K)
for(j in 1:ell) KK[,j] <- dnorm(xx, omega[j], sd=0.5) 
p <- predict(fit, newdata=KK, interval="prediction")
plot(x, y)
lines(xx, p[,1])
lines(xx, p[,2], col=2, lty=2)
lines(xx, p[,3], col=2, lty=2)

ell <- 20
omega <- maximinLHS(ell, 2)
omega[,1] <- (omega[,1] - 0.5)*6 + 1
omega[,2] <- (omega[,2] - 0.5)*6 + 1

n <- nrow(X2)
K <- as.data.frame(matrix(NA, ncol=ell, nrow=n))
for(j in 1:ell) 
  K[,j] <- dnorm(X2[,1], omega[j,1])*dnorm(X2[,2], omega[j,2])
names(K) <- paste0("omega", 1:ell)

fit <- lm(y2 ~ . -1, data=K)

xx <- seq(-2, 4, length=gn)
XX <- expand.grid(xx, xx)
KK <- as.data.frame(matrix(NA, ncol=ell, nrow=nrow(XX)))
for(j in 1:ell) 
  KK[,j] <- dnorm(XX[,1], omega[j,1])*dnorm(XX[,2], omega[j,2])
names(KK) <- paste0("omega", 1:ell)

p <- predict(fit, newdata=KK, se.fit=TRUE)

par(mfrow=c(1,2))
image(xx, xx, matrix(p$fit, ncol=gn), col=cols, main="mean", 
      xlab="x1", ylab="x2")
points(X2)
points(omega, pch=20)
image(xx, xx, matrix(p$se.fit, ncol=gn), col=cols, main="sd", 
      xlab="x1", ylab="x2")
points(X2)
points(omega, pch=20)
par(mfrow=c(1,1))

#stationarity
X <- seq(0,20,length=100)
y <- (sin(pi*X/5) + 0.2*cos(4*pi*X/5)) * (X <= 9.6)
lin <- X>9.6
y[lin] <- -1 + X[lin]/10
y <- y + rnorm(length(y), sd=0.1)

gpi <- newGP(matrix(X, ncol=1), y, d=0.1, g=0.1*var(y), dK=TRUE)
mle <- jmleGP(gpi)

p <- predGP(gpi, matrix(X, ncol=1), lite=TRUE)
deleteGP(gpi)

plot(X, y, xlab="x")
lines(X, p$mean)
lines(X, p$mean + 2*sqrt(p$s2), col=2, lty=2)
lines(X, p$mean - 2*sqrt(p$s2), col=2, lty=2)

left <- X < 9.6
gpl <- newGP(matrix(X[left], ncol=1), y[left], d=0.1, 
             g=0.1*var(y), dK=TRUE)
mlel <- jmleGP(gpl)
gpr <- newGP(matrix(X[!left], ncol=1), y[!left], d=0.1, 
             g=0.1*var(y), dK=TRUE)
mler <- jmleGP(gpr, drange=c(eps, 100)) 

pl <- predGP(gpl, matrix(X[left], ncol=1), lite=TRUE)
deleteGP(gpl)
pr <- predGP(gpr, matrix(X[!left], ncol=1), lite=TRUE)
deleteGP(gpr)

plot(X, y, xlab="x")
lines(X[left], pl$mean)
lines(X[left], pl$mean + 2*sqrt(pl$s2), col=2, lty=2)
lines(X[left], pl$mean - 2*sqrt(pl$s2), col=2, lty=2)
lines(X[!left], pr$mean)
lines(X[!left], pr$mean + 2*sqrt(pr$s2), col=2, lty=2)
lines(X[!left], pr$mean - 2*sqrt(pr$s2), col=2, lty=2)

#packages
# Install and load the gplm package
install.packages("gplm")
library(gplm)

# Fit a Gaussian Process model to predict 'mpg' using 'hp' and 'wt'
data <- mtcars[, c("mpg", "hp", "wt")]
fit <- gplm(y = data$mpg, X = data[, c("hp", "wt")], kern.type = "rbf")

# Summary of the fit
summary(fit)

# Predictions
newdata <- data.frame(hp = c(110, 150), wt = c(2.5, 3.5))
predict(fit, Xnew = newdata)

# Install and load the DiceKriging package
install.packages("DiceKriging")
library(DiceKriging)

# Generate a simple spatial grid
set.seed(123)
x <- seq(-5, 5, length = 20)
y <- seq(-5, 5, length = 20)
grid <- expand.grid(x = x, y = y)

# Create a response variable with a spatial pattern
z <- with(grid, sin(sqrt(x^2 + y^2)))

# Convert to a data frame
spatial_data <- data.frame(grid, z)

# Fit the Gaussian Process model using DiceKriging
gp_model <- km(
  formula = z ~ 1,                   # Response variable
  design = spatial_data[, c("x", "y")],  # Predictor variables (spatial coordinates)
  response = spatial_data$z,            # Response variable
  covtype = "gauss",                    # Gaussian covariance function (kernel)
  nugget = 1e-5,                        # Small nugget effect for numerical stability
  control = list(trace = TRUE)
)

summary(gp_model)

# Generate a grid for predictions
pred_grid <- expand.grid(x = seq(-5, 5, length = 40), y = seq(-5, 5, length = 40))

# Predict using the fitted Gaussian Process model
# Type "UK" is used for Universal Kriging (most common choice)
predictions <- predict(gp_model, newdata = pred_grid, type = "UK")

# Add predictions to the grid for easy plotting
pred_grid$predicted_z <- predictions$mean

# Original data plot
library(ggplot2)
ggplot(spatial_data, aes(x = x, y = y, fill = z)) +
  geom_tile() +
  labs(title = "Original Spatial Data", fill = "z")

# Predicted data plot
ggplot(pred_grid, aes(x = x, y = y, fill = predicted_z)) +
  geom_tile() +
  labs(title = "Predicted Spatial Data using DiceKriging", fill = "Predicted z")

# Install and load the kernlab package
install.packages("kernlab")
library(kernlab)

# Fit a Gaussian Process model to predict 'mpg' using 'hp' and 'wt'
data <- as.matrix(mtcars[, c("hp", "wt")])
gp_fit <- gausspr(x = data, y = mtcars$mpg, kernel = "rbfdot", kpar = list(sigma = 0.1))

# Summary of the fit
print(gp_fit)

# Predictions
newdata <- as.matrix(data.frame(hp = c(110, 150), wt = c(2.5, 3.5)))
predict(gp_fit, newdata)

# Install and load the laGP package
install.packages("laGP")
library(laGP)

# Load necessary libraries
install.packages("laGP")
library(laGP)

# Example dataset (using mtcars with mpg as the response variable and a few predictors)
X <- as.matrix(mtcars[, c("disp", "hp", "wt", "qsec")])  # Predictor variables
y <- mtcars$mpg  # Response variable

# Set hyperparameters for the GP
d <- darg(list(mle = TRUE), X)  # Automatically estimate length-scale (d)
g <- garg(list(mle = TRUE), y)  # Automatically estimate nugget (g)

# Set the number of initial points (start) for the laGP method
start <- 10  # You can choose a reasonable value based on your dataset size
end <- 20    # The total number of points to include in the final model

# Run the laGP function
result <- laGP(Xref = X, X = X, Z = y, d = d$start, g = g$start, method = "alc", start = start, end = end)

# Predictions and other results can be extracted from the result object
predictions <- result$mean

# Print predictions
print(predictions)

# Install and load the RobustGaSP package
install.packages("RobustGaSP")
library(RobustGaSP)

# Select variables from mtcars
X <- as.matrix(mtcars[, c("hp", "wt")])
y <- mtcars$mpg

# Fit a Robust Gaussian Process model
rgp_model <- rgasp(design = X, response = y, kernel_type = "matern_5_2")

# Summary of the model
summary(rgp_model)

# Predictions for new data points
new_X <- as.matrix(data.frame(hp = c(110, 150), wt = c(2.5, 3.5)))
predictions <- predict(rgp_model, testing_input = new_X)

print(predictions$mean)  # Predicted means

