# 10/8 CSC 401 Examples

# Load necessary packages
library(MASS)  # For stepAIC if needed

# Load the mtcars dataset
data(mtcars)

# Define the response variable and predictor variables
response <- "mpg"
predictors <- setdiff(names(mtcars), response)

# Function to calculate Adjusted R-squared
adj_r_squared <- function(model) {
  r2 <- summary(model)$r.squared
  n <- nrow(model$model)
  p <- length(model$coefficients) - 1
  adj_r2 <- 1 - (1 - r2) * (n - 1) / (n - p - 1)
  return(adj_r2)
}

# Function to fit models for each subset of predictors and evaluate
best_subset_selection <- function(data, response, predictors) {
  best_model <- NULL
  best_adj_r2 <- -Inf
  best_subset <- NULL
  
  # Generate all possible subsets of predictors
  for (k in 1:length(predictors)) {
    subsets <- combn(predictors, k, simplify = FALSE)
    
    for (subset in subsets) {
      formula <- as.formula(paste(response, "~", paste(subset, collapse = " + ")))
      model <- lm(formula, data = data)
      
      # Calculate Adjusted R-squared
      current_adj_r2 <- adj_r_squared(model)
      
      if (current_adj_r2 > best_adj_r2) {
        best_adj_r2 <- current_adj_r2
        best_model <- model
        best_subset <- subset
      }
    }
  }
  
  return(list(model = best_model, subset = best_subset, adj_r2 = best_adj_r2))
}

# Apply best subset selection
result <- best_subset_selection(mtcars, response, predictors)

# Print the results
cat("Best Subset:\n")
print(result$subset)
cat("\nBest Model Summary:\n")
print(summary(result$model))
cat("\nAdjusted R-squared:", result$adj_r2, "\n")

#backward selection
# Load necessary packages
library(MASS)  # For stepAIC if needed

# Load the mtcars dataset
data(mtcars)

# Define the response variable and predictor variables
response <- "mpg"
predictors <- setdiff(names(mtcars), response)

# Function to fit the model and get summary statistics
fit_model <- function(data, response, predictors) {
  formula <- as.formula(paste(response, "~", paste(predictors, collapse = " + ")))
  model <- lm(formula, data = data)
  return(model)
}

# Function to get the highest p-value from model summary
get_highest_pvalue <- function(model) {
  summary(model)$coefficients[, "Pr(>|t|)"][-1]  # Exclude the intercept
}

# Function to perform backward selection
backward_selection <- function(data, response, predictors, threshold = 0.05) {
  current_predictors <- predictors
  repeat {
    # Fit the model with the current predictors
    model <- fit_model(data, response, current_predictors)
    
    # Get p-values of the predictors
    pvalues <- get_highest_pvalue(model)
    
    # Check if any p-value is greater than the threshold
    max_pvalue <- max(pvalues, na.rm = TRUE)
    if (max_pvalue <= threshold) {
      break
    }
    
    # Find the predictor with the highest p-value
    predictor_to_remove <- names(pvalues)[which.max(pvalues)]
    
    # Remove this predictor from the current list
    current_predictors <- setdiff(current_predictors, predictor_to_remove)
  }
  
  # Final model
  final_model <- fit_model(data, response, current_predictors)
  return(list(model = final_model, predictors = current_predictors))
}

# Function to calculate regression metrics
regression_metrics <- function(model) {
  residuals <- model$residuals
  fitted_values <- model$fitted.values
  n <- length(residuals)
  p <- length(model$coefficients) - 1
  
  # Calculate metrics
  sse <- sum(residuals^2)  # Sum of Squared Errors
  sst <- sum((mtcars[[response]] - mean(mtcars[[response]]))^2)  # Total Sum of Squares
  r_squared <- 1 - sse / sst  # R-squared
  adj_r_squared <- 1 - (1 - r_squared) * (n - 1) / (n - p - 1)  # Adjusted R-squared
  rmse <- sqrt(sse / n)  # Root Mean Squared Error
  mape <- mean(abs(residuals / mtcars[[response]])) * 100  # Mean Absolute Percentage Error
  aic <- AIC(model)  # Akaike Information Criterion
  bic <- BIC(model)  # Bayesian Information Criterion
  
  return(list(
    R_squared = r_squared,
    Adjusted_R_squared = adj_r_squared,
    RMSE = rmse,
    MAPE = mape,
    AIC = aic,
    BIC = bic
  ))
}

# Apply backward selection
result <- backward_selection(mtcars, response, predictors)

# Print the results
cat("Final Model Summary:\n")
print(summary(result$model))

cat("\nSelected Predictors:\n")
print(result$predictors)

cat("\nRegression Metrics:\n")
metrics <- regression_metrics(result$model)
print(metrics)

#model selection, parametric and nonparametric on one variable
# Load necessary libraries
library(mgcv)    # For GAMs and penalized splines
library(splines) # For B-splines
library(kernlab) # For Gaussian Process

# Define a function to calculate MAPE
mape <- function(actual, predicted) {
  mean(abs((actual - predicted) / actual)) * 100
}

# Define a function to compare models
compare_models <- function(data, response, predictor) {
  # Extract response and predictor
  y <- data[[response]]
  x <- data[[predictor]]
  
  # Prepare data
  model_data <- data.frame(x = x, y = y)
  
  # Linear model
  linear_model <- lm(y ~ x, data = model_data)
  linear_pred <- predict(linear_model, newdata = model_data)
  linear_mape <- mape(y, linear_pred)
  
  # Quadratic model
  quadratic_model <- lm(y ~ x + I(x^2), data = model_data)
  quadratic_pred <- predict(quadratic_model, newdata = model_data)
  quadratic_mape <- mape(y, quadratic_pred)
  
  # Cubic model
  cubic_model <- lm(y ~ x + I(x^2) + I(x^3), data = model_data)
  cubic_pred <- predict(cubic_model, newdata = model_data)
  cubic_mape <- mape(y, cubic_pred)
  
  # Quartic model
  quartic_model <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4), data = model_data)
  quartic_pred <- predict(quartic_model, newdata = model_data)
  quartic_mape <- mape(y, quartic_pred)
  
  # LOESS model
  loess_model <- loess(y ~ x, data = model_data)
  loess_pred <- predict(loess_model, newdata = model_data)
  loess_mape <- mape(y, loess_pred)
  
  # Smoothing spline model
  smooth_spline_model <- smooth.spline(x, y)
  spline_pred <- predict(smooth_spline_model, x)$y
  spline_mape <- mape(y, spline_pred)
  
  # Penalized B-spline model
  penalty_spline_model <- gam(y ~ s(x, bs = "cs"), data = model_data)
  penalty_spline_pred <- predict(penalty_spline_model, newdata = model_data)
  penalty_spline_mape <- mape(y, penalty_spline_pred)
  
  # Gaussian Process model
  gp_model <- gausspr(x = matrix(x), y = y, kernel = rbfdot(sigma = 0.1))
  gp_pred <- predict(gp_model, matrix(x))
  gp_mape <- mape(y, gp_pred)
  
  # Collect results
  results <- data.frame(
    Model = c("Linear", "Quadratic", "Cubic", "Quartic", "LOESS", "Smoothing Spline", "Penalized B-spline", "Gaussian Process"),
    MAPE = c(linear_mape, quadratic_mape, cubic_mape, quartic_mape, loess_mape, spline_mape, penalty_spline_mape, gp_mape)
  )
  
  # Find the best model
  best_model <- results[which.min(results$MAPE), ]
  
  return(list(
    results = results,
    best_model = best_model
  ))
}

# Example usage with mtcars dataset
model_comparison <- compare_models(mtcars, "mpg", "hp")

# Print results
print(model_comparison$results)
print(paste("Best model based on MAPE:", model_comparison$best_model$Model))

#with plotting results
# Load necessary libraries
library(mgcv)    # For GAMs and penalized splines
library(splines) # For B-splines
library(kernlab) # For Gaussian Process
library(ggplot2) # For plotting

# Define a function to calculate MAPE
mape <- function(actual, predicted) {
  mean(abs((actual - predicted) / actual)) * 100
}

# Define a function to compare models and plot results
compare_models <- function(data, response, predictor) {
  # Extract response and predictor
  y <- data[[response]]
  x <- data[[predictor]]
  
  # Prepare data
  model_data <- data.frame(x = x, y = y)
  
  # Linear model
  linear_model <- lm(y ~ x, data = model_data)
  linear_pred <- predict(linear_model, newdata = model_data)
  linear_mape <- mape(y, linear_pred)
  
  # Quadratic model
  quadratic_model <- lm(y ~ x + I(x^2), data = model_data)
  quadratic_pred <- predict(quadratic_model, newdata = model_data)
  quadratic_mape <- mape(y, quadratic_pred)
  
  # Cubic model
  cubic_model <- lm(y ~ x + I(x^2) + I(x^3), data = model_data)
  cubic_pred <- predict(cubic_model, newdata = model_data)
  cubic_mape <- mape(y, cubic_pred)
  
  # Quartic model
  quartic_model <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4), data = model_data)
  quartic_pred <- predict(quartic_model, newdata = model_data)
  quartic_mape <- mape(y, quartic_pred)
  
  # LOESS model
  loess_model <- loess(y ~ x, data = model_data)
  loess_pred <- predict(loess_model, newdata = model_data)
  loess_mape <- mape(y, loess_pred)
  
  # Smoothing spline model
  smooth_spline_model <- smooth.spline(x, y)
  spline_pred <- predict(smooth_spline_model, x)$y
  spline_mape <- mape(y, spline_pred)
  
  # Penalized B-spline model
  penalty_spline_model <- gam(y ~ s(x, bs = "cs"), data = model_data)
  penalty_spline_pred <- predict(penalty_spline_model, newdata = model_data)
  penalty_spline_mape <- mape(y, penalty_spline_pred)
  
  # Gaussian Process model
  gp_model <- gausspr(x = matrix(x), y = y, kernel = rbfdot(sigma = 0.1))
  gp_pred <- predict(gp_model, matrix(x))
  gp_mape <- mape(y, gp_pred)
  
  # Collect results
  results <- data.frame(
    Model = c("Linear", "Quadratic", "Cubic", "Quartic", "LOESS", "Smoothing Spline", "Penalized B-spline", "Gaussian Process"),
    MAPE = c(linear_mape, quadratic_mape, cubic_mape, quartic_mape, loess_mape, spline_mape, penalty_spline_mape, gp_mape)
  )
  
  # Find the best model
  best_model <- results[which.min(results$MAPE), ]
  
  # Plot results
  p <- ggplot(results, aes(x = reorder(Model, MAPE), y = MAPE)) +
    geom_bar(stat = "identity", fill = "steelblue") +
    coord_flip() +
    labs(title = "Model Comparison by MAPE", x = "Model", y = "MAPE") +
    theme_minimal()
  
  # Print plot
  print(p)
  
  return(list(
    results = results,
    best_model = best_model
  ))
}

# Example usage with mtcars dataset
model_comparison <- compare_models(mtcars, "mpg", "hp")

# Print results
print(model_comparison$results)
print(paste("Best model based on MAPE:", model_comparison$best_model$Model))

#forward selection to add quadratic and interaction terms to end of backward selection model
# Load necessary libraries
library(dplyr)

# Define the initial model with variables selected through backward selection
initial_vars <- c("wt", "qsec", "am")
data <- mtcars

# Start with the initial model
model_formula <- as.formula(paste("mpg ~", paste(initial_vars, collapse = " + ")))
current_model <- lm(model_formula, data = data)

# Function to add polynomial and interaction terms
add_polynomial_and_interaction_terms <- function(data, initial_vars, current_model) {
  new_vars <- initial_vars
  max_degree <- 2
  
  # Consider adding polynomial terms
  for (var in initial_vars) {
    for (degree in 2:max_degree) {
      new_term <- paste0("I(", var, "^", degree, ")")
      model_formula <- as.formula(paste("mpg ~", paste(c(new_vars, new_term), collapse = " + ")))
      new_model <- lm(model_formula, data = data)
      
      # Check if the term is present in the coefficients
      if (new_term %in% rownames(summary(new_model)$coefficients)) {
        p_value <- summary(new_model)$coefficients[new_term, 4]
        
        if (p_value < 0.05) {
          new_vars <- c(new_vars, new_term)
          current_model <- new_model
        }
      }
    }
  }
  
  # Consider adding interaction terms
  interaction_combinations <- combn(initial_vars, 2, simplify = FALSE)
  for (interaction in interaction_combinations) {
    new_term <- paste(interaction, collapse = ":")
    model_formula <- as.formula(paste("mpg ~", paste(c(new_vars, new_term), collapse = " + ")))
    new_model <- lm(model_formula, data = data)
    
    # Check if the term is present in the coefficients
    if (new_term %in% rownames(summary(new_model)$coefficients)) {
      p_value <- summary(new_model)$coefficients[new_term, 4]
      
      if (p_value < 0.05) {
        new_vars <- c(new_vars, new_term)
        current_model <- new_model
      }
    }
  }
  
  return(current_model)
}

# Apply the function to add polynomial and interaction terms
final_model <- add_polynomial_and_interaction_terms(data, initial_vars, current_model)

# Output the final model summary
summary(final_model)

# Calculate additional metrics
rsq <- summary(final_model)$r.squared
adj_rsq <- summary(final_model)$adj.r.squared
aic_value <- AIC(final_model)
bic_value <- BIC(final_model)
rmse <- sqrt(mean(residuals(final_model)^2))

cat("R-squared: ", rsq, "\n")
cat("Adjusted R-squared: ", adj_rsq, "\n")
cat("AIC: ", aic_value, "\n")
cat("BIC: ", bic_value, "\n")
cat("RMSE: ", rmse, "\n")
