# 11/12 CSC 401 Examples

#factor analysis
library(psych)
fa_result <- fa(mtcars, nfactors = 2, rotate = "varimax")
print(fa_result)

library(FactoMineR)
pca_result <- PCA(mtcars, graph = FALSE)
plot(pca_result)

# Convert some variables to factors to create mixed data
mtcars$am <- as.factor(mtcars$am)
mtcars$cyl <- as.factor(mtcars$cyl)

# Perform FAMD
library(FactoMineR)
famd_result <- FAMD(mtcars, ncp = 5)
plot(famd_result)


library(lavaan)

# Use built-in dataset for demonstration
data("HolzingerSwineford1939")

# Model specification
model <- '
   # latent variables
   visual  =~ x1 + x2 + x3
   textual =~ x4 + x5 + x6
   speed   =~ x7 + x8 + x9
'

# Fit the model
fit <- lavaan::sem(model, data = HolzingerSwineford1939)

# Summary of the model
summary(fit, fit.measures = TRUE)

#PCA
data(mtcars)
pca_result <- prcomp(mtcars, scale. = TRUE)
summary(pca_result)

library(FactoMineR)
pca_result <- PCA(mtcars, graph = FALSE)
plot(pca_result)

library(factoextra)
fviz_pca_biplot(pca_result)


#PCR
library(pls)
pcr_model <- pcr(mpg ~ ., data = mtcars, scale = TRUE, validation = "CV")
summary(pcr_model)
plot(pcr_model, validationplot = TRUE)

library(caret)
pcr_model <- train(mpg ~ ., data = mtcars, method = "pcr", preProcess = "scale", trControl = trainControl(method = "cv"))
summary(pcr_model)

# Load necessary libraries
library(ggplot2)

# Perform PCA on the mtcars dataset
pca_result <- prcomp(mtcars, scale. = TRUE)

# Calculate the proportion of variance explained
explained_variance <- pca_result$sdev^2 / sum(pca_result$sdev^2)

# Create a scree plot
scree_data <- data.frame(Principal_Component = seq_along(explained_variance),
                         Variance_Explained = explained_variance)

ggplot(scree_data, aes(x = Principal_Component, y = Variance_Explained)) +
  geom_point(size = 4) +
  geom_line() +
  scale_x_continuous(breaks = scree_data$Principal_Component) +
  xlab("Principal Component") +
  ylab("Proportion of Variance Explained") +
  ggtitle("Scree Plot for PCA on mtcars Dataset") +
  theme_minimal()



#factor analysis algorithm
scores <- read.table("https://raw.githubusercontent.com/sunbeomk/PSYC490/main/scores.txt")
head(scores)
cor_subtests <- cor(scores)

round(cor_subtests, 2)
efa_1 <- factanal(x = scores, factors = 1)
class(efa_1)
efa_1
loadings <- efa_1$loadings # Factor loadings
loadings[1]
as.numeric(loadings)
sum(loadings^2) # SS Loadings
sum(loadings^2) / ncol(scores) # Proportion Var
factanal(x = scores, 
         factors = 2)

cov_mat <- cov(scores)
round(cov_mat, 2)
factanal(covmat = cov_mat, 
         factors = 2, 
         n.obs = nrow(scores))
output <- factanal(x = scores, 
                   factors = 2, 
                   scores = "regression")
head(output$scores)

varimax <- factanal(scores, 
                    factors = 2, 
                    rotation="varimax", 
                    scores="regression")
cor(varimax$scores)

promax <- factanal(scores, 
                   factors = 2, 
                   rotation = "promax", 
                   scores = "regression")
cor(promax$scores)

library(psych)
output2 <- fa(scores, # input data
              nfactors = 2, # number of factors
              rotate = "varimax", # rotation
              scores = "regression") # factor score estimation

output2$loadings # factor loadings
output2$uniquenesses # uniqueness
output2$communality # communality
output2$uniquenesses + output2$communalities

# Load necessary libraries
library(ggplot2)
library(psych)

# Load the dataset
scores <- read.table("https://raw.githubusercontent.com/sunbeomk/PSYC490/main/scores.txt", header = TRUE)

# Compute the correlation matrix
cor_matrix <- cor(scores)

# Plot the correlation matrix heatmap
ggplot(data = as.data.frame(as.table(cor_matrix)), aes(Var1, Var2, fill = Freq)) +
  geom_tile() + 
  scale_fill_gradient2(low = "blue", high = "red", mid = "white", midpoint = 0) +
  theme_minimal() +
  labs(title = "Correlation Matrix", x = "", y = "")

# Perform PCA
pca <- prcomp(scores, scale. = TRUE)

# Scree Plot
screeplot(pca, type = "lines", main = "Scree Plot")

# Get PCA loadings
loadings <- pca$rotation

# Plot the loadings for the first principal component
barplot(loadings[,1], main = "PCA Loadings for 1st Component", ylab = "Loading", names.arg = colnames(scores))

# Perform Multidimensional Scaling (MDS)
mds <- cmdscale(dist(scores))

# Scatterplot of MDS results
plot(mds, xlab = "Dimension 1", ylab = "Dimension 2", main = "MDS Scatterplot of Factor Scores", pch = 16, col = "blue")


#PCA
# Load necessary libraries
library(mlbench)
library(ggplot2)

# Load the BostonHousing dataset
data("mtcars")

# Extract the numeric variables (excluding the response variable 'mpg')
data_numeric <- mtcars[, -which(names(mtcars) == "mpg")]

# Standardize the data
data_standardized <- scale(data_numeric)

# Calculate the covariance matrix
cov_matrix <- cov(data_standardized)

# Perform eigen decomposition
eigen_decomp <- eigen(cov_matrix)
eigenvalues <- eigen_decomp$values
eigenvectors <- eigen_decomp$vectors

# Calculate the explained variance
explained_variance <- eigenvalues / sum(eigenvalues)

# Project the data onto the principal components
pca_scores <- data_standardized %*% eigenvectors
colnames(pca_scores) <- paste0("PC", 1:ncol(pca_scores))

# Create a scree plot
explained_variance_df <- data.frame(
  PC = paste0("PC", 1:length(explained_variance)),
  Variance_Explained = explained_variance
)

ggplot(explained_variance_df, aes(x = PC, y = Variance_Explained)) +
  geom_bar(stat = "identity", fill = "steelblue") +
  geom_point(size = 3, color = "red") +
  geom_line(group = 1, color = "red") +
  labs(title = "Scree Plot", x = "Principal Component", y = "Variance Explained") +
  theme_minimal()

# Loadings plot
loadings_df <- as.data.frame(eigenvectors[, 1:2])
colnames(loadings_df) <- c("PC1", "PC2")
loadings_df$Variable <- rownames(loadings_df)

ggplot(loadings_df, aes(x = PC1, y = PC2, label = Variable)) +
  geom_point() +
  geom_text(vjust = 1.5, hjust = 1.5) +
  labs(title = "Loadings Plot", x = "PC1", y = "PC2") +
  theme_minimal()

# Biplot
biplot_df <- data.frame(pca_scores[, 1:2])
biplot_df$Label <- rownames(mtcars)

ggplot(biplot_df, aes(x = PC1, y = PC2)) +
  geom_point(aes(color = Label), alpha = 0.6) +
  geom_text(aes(label = Label), vjust = 1.5, hjust = 1.5, size = 2) +
  labs(title = "PCA Biplot", x = "PC1", y = "PC2") +
  theme_minimal()

#PCR
# Load necessary libraries
library(ggplot2)

# Use the mtcars dataset
data <- mtcars
target <- data$mpg
data_numeric <- data[, -which(names(data) == "mpg")]

# Standardize the data
data_standardized <- scale(data_numeric)

# Calculate the covariance matrix
cov_matrix <- cov(data_standardized)

# Perform eigen decomposition
eigen_decomp <- eigen(cov_matrix)
eigenvalues <- eigen_decomp$values
eigenvectors <- eigen_decomp$vectors

# Project the data onto the principal components
pca_scores <- data_standardized %*% eigenvectors
colnames(pca_scores) <- paste0("PC", 1:ncol(pca_scores))

# Fit a linear model using the principal components
pcr_model <- lm(target ~ pca_scores)

# Summary of the PCR model
summary(pcr_model)

# Extract the coefficients
coefficients_df <- data.frame(
  PC = paste0("PC", 1:ncol(pca_scores)),
  Coefficients = pcr_model$coefficients[-1]  # Exclude the intercept
)

# Plot the coefficients
ggplot(coefficients_df, aes(x = PC, y = Coefficients)) +
  geom_bar(stat = "identity", fill = "steelblue") +
  labs(title = "PCR Coefficients", x = "Principal Component", y = "Coefficient Value") +
  theme_minimal()

# Predict the values using the PCR model
predicted_values <- predict(pcr_model)

# Plot the predicted vs actual values
pred_vs_actual <- data.frame(
  Actual = target,
  Predicted = predicted_values
)

ggplot(pred_vs_actual, aes(x = Actual, y = Predicted)) +
  geom_point(color = "blue") +
  geom_abline(slope = 1, intercept = 0, linetype = "dashed", color = "red") +
  labs(title = "Predicted vs Actual MPG", x = "Actual MPG", y = "Predicted MPG") +
  theme_minimal()

# Use only the first three principal components
selected_pcs <- pca_scores[, 1:3]

# Fit a linear model using the first three principal components
pcr_model_selected <- lm(target ~ selected_pcs)

# Summary of the new PCR model
summary(pcr_model_selected)

# Extract the coefficients for the selected components
coefficients_df_selected <- data.frame(
  PC = colnames(selected_pcs),
  Coefficients = pcr_model_selected$coefficients[-1]  # Exclude the intercept
)

# Plot the coefficients
ggplot(coefficients_df_selected, aes(x = PC, y = Coefficients)) +
  geom_bar(stat = "identity", fill = "steelblue") +
  labs(title = "PCR Coefficients (First 3 PCs)", x = "Principal Component", y = "Coefficient Value") +
  theme_minimal()

# Predict the values using the selected PCR model
predicted_values_selected <- predict(pcr_model_selected)

# Plot the predicted vs actual values
pred_vs_actual_selected <- data.frame(
  Actual = target,
  Predicted = predicted_values_selected
)

ggplot(pred_vs_actual_selected, aes(x = Actual, y = Predicted)) +
  geom_point(color = "blue") +
  geom_abline(slope = 1, intercept = 0, linetype = "dashed", color = "red") +
  labs(title = "Predicted vs Actual MPG (First 3 PCs)", x = "Actual MPG", y = "Predicted MPG") +
  theme_minimal()

