# 11/19 CSC 401 Examples

# Load necessary library
library(kernlab)

# Load iris dataset (using only numeric variables)
data(iris)
iris_data <- iris[, -5]  # Remove the species column for clustering

# Check the dimensions of the data to ensure correctness
print(dim(iris_data))  # Should return 150 rows and 4 columns

# Apply spectral clustering
# Ensure that centers parameter is set correctly
sc <- specc(as.matrix(iris_data), centers = 3)

# Check the length of the clustering result
length(sc)  # This should return 150, not 4

# Check the length of the true labels
length(iris$Species)  # Should be 150

# If the lengths match, create the table
if(length(sc) == length(iris$Species)) {
  clustering_results <- table(sc, iris$Species)
  print(clustering_results)
} else {
  stop("The lengths of the clustering results and true labels do not match.")
}

# Assuming all above steps are correct
# Apply spectral clustering again
sc <- specc(as.matrix(iris_data), centers = 3)

# Check the output
print(length(sc))  # Should be 150

# Generate confusion matrix if lengths are correct
if (length(sc) == 150) {
  clustering_results <- table(sc, iris$Species)
  print(clustering_results)
} else {
  print("The lengths of the clustering results and the true labels do not match.")
}

# Visualization
plot(iris_data, col = as.factor(sc), pch = 19, main = "Spectral Clustering on Iris Dataset")

#scratch
# Load the iris dataset, excluding the species column
data(iris)
iris_data <- iris[, -5]  # Use only the numeric columns

# Set a sigma value for the Gaussian kernel
sigma <- 1

# Construct the similarity matrix using Gaussian kernel
n <- nrow(iris_data)
similarity_matrix <- matrix(0, n, n)
for (i in 1:n) {
  for (j in 1:n) {
    similarity_matrix[i, j] <- exp(-sum((iris_data[i, ] - iris_data[j, ])^2) / (2 * sigma^2))
  }
}

# View the similarity matrix
head(similarity_matrix)

# Compute the degree matrix
degree_matrix <- diag(rowSums(similarity_matrix))

# Compute the Laplacian matrix
laplacian_matrix <- degree_matrix - similarity_matrix

# View the Laplacian matrix
head(laplacian_matrix)

# Compute eigenvalues and eigenvectors
eigen_decomp <- eigen(laplacian_matrix)
eigenvalues <- eigen_decomp$values
eigenvectors <- eigen_decomp$vectors

# Select the eigenvectors corresponding to the smallest non-zero eigenvalues
# Let's select the first 2 or 3 eigenvectors, depending on the clusters you want
selected_eigenvectors <- eigenvectors[, (n-2):(n-1)]  # Selects the last 2 eigenvectors

# View the selected eigenvectors
head(selected_eigenvectors)

# Apply k-means clustering to the selected eigenvectors
set.seed(123)  # For reproducibility
k <- 3  # Number of clusters
clusters <- kmeans(selected_eigenvectors, centers = k)$cluster

# Compare the resulting clusters with the true species labels
table(clusters, iris$Species)

# Visualize the clusters
plot(iris_data[, 1:2], col = clusters, pch = 19, 
     main = "Spectral Clustering on Iris Dataset")

# Plot the rows of the selected eigenvectors
plot(selected_eigenvectors, col = clusters, pch = 19,
     main = "Clustering in the Eigenvector Space")
