# 11/26 CSC 401 Examples

#simple nn
# for manipulating data
library(dplyr)
# for parallel processing
library(purrr)
# for neural network functionality, install first
library(torch)

url_data <- "https://raw.githubusercontent.com/mikeyEcology/Dada-Analysis-on-Diamonds-Dataset/main/diamonds.csv"

raw_data <- url_data |>
  url() |>
  vroom::vroom()
continuous_features <- c("carat", "depth", "table", "x", "y", "z")
target_name <- "price"

features <- raw_data |>
  dplyr::select(dplyr::all_of(continuous_features)) |>
  scale() |>
  # scale converts the data into an unfortunate format. Make it back to df
  tibble::as_tibble()

target <- raw_data |>
  dplyr::select(dplyr::all_of(target_name)) |>
  scale() |>
  tibble::as_tibble()

train_val <- dplyr::bind_cols(features, target)

apply(train_val, 2, mean)

apply(train_val, 2, sd)

pairs(train_val)

dset <- torch::dataset(
  name = "dset",
  
  initialize = function(indices) {
    data <- self$prepare_data(train_val[indices, ])
    self$x <- data[[1]]
    self$y <- data[[2]]
  },
  
  .getitem = function(i) {
    x <- self$x[i, ]
    y <- self$y[i, ]
    
    list(x, y)
  },
  
  .length = function() {
    dim(self$y)[1]
  },
  
  prepare_data = function(input) {
    feature_cols <- input |>
      dplyr::select(dplyr::all_of(continuous_features)) |>
      as.matrix()
    
    target_col <- input |>
      dplyr::select(dplyr::all_of(target_name)) |>
      as.matrix()
    
    list(
      torch_tensor(feature_cols),
      torch_tensor(target_col)
    )
  }
)

train_indices <- sample(1:nrow(train_val), size = floor(0.8 * nrow(train_val)))
valid_indices <- setdiff(1:nrow(train_val), train_indices)

# bigger batch sizes train faster. General concensus is that this is not a hyperparameter worth tuning. 
batch_size = 256

train_ds <- dset(train_indices)
train_dl <- train_ds |>
  dataloader(batch_size = batch_size, shuffle = TRUE)

valid_ds <- dset(valid_indices)
valid_dl <- valid_ds |>
  dataloader(batch_size = batch_size, shuffle = FALSE)

net <- nn_module(
  "net",
  
  initialize = function(num_numerical, fc1_dim
  ) {
    self$fc1 <- nn_linear(num_numerical, fc1_dim)
    self$output <- nn_linear(fc1_dim, 1)
  },
  
  forward = function(x) {
    x |> 
      self$fc1() |>
      nnf_relu() |>
      self$output() |>
      nnf_sigmoid()
  }
)

# This is the number of neurons in the hidden layer of the neural network
fc1_dim <- 32
# The number of features is the number of neurons in the input layer
num_numerical <- length(continuous_features)

# build this neural net
model <- net(num_numerical, fc1_dim)

device <- if (cuda_is_available()) torch_device("cuda:0") else "cpu"

model <- model$to(device = device)

learning_rate <- 0.01
# using stochastic gradient descent for the optimizer
optimizer <- optim_sgd(model$parameters, lr = learning_rate, momentum = 0)
num_epochs <- 10
# using mean squared error loss
loss_func <- nnf_mse_loss

for (epoch in 1:num_epochs) {
  
  model$train()
  train_losses <- c()  
  
  coro::loop(for (b in train_dl) {
    optimizer$zero_grad()
    output <- model(b[[1]]$to(device = device))
    loss <- loss_func(output, b[[2]]$to(dtype = torch_float(), device = device))
    loss$backward()
    optimizer$step()
    train_losses <- c(train_losses, loss$item())
  })
  
  model$eval()
  valid_losses <- c()
  
  coro::loop(for (b in valid_dl) {
    output <- model(b[[1]]$to(device = device))
    loss <- loss_func(output, b[[2]]$to(dtype = torch_float(), device = device))
    valid_losses <- c(valid_losses, loss$item())
  })
  
  cat(sprintf("Epoch %d: training loss: %3f, validation loss: %3f\n", epoch, mean(train_losses), mean(valid_losses)))
}

# we set model to evaluation model so it knows not to calculate gradients now
model$eval()

# predict the target for the validation dataset
target_pred <- c()
coro::loop(for (b in valid_dl) {
  output <- model(b[[1]]$to(device = device))
  #target_pred <- c(target_pred, output)
  for (i in 1:length(output)) {
    pred_array <- as_array(output)[i,]
    target_pred <- c(target_pred, pred_array)
  }
})

# get the observed target from the validation dataset
# we use `as_array` to convert from tensors to standard R data structures
target_obs <- as_array(valid_ds$y)

# get mean and std dev from original data (these were used when scaling by the `scale` function)
raw_target <- raw_data |>
  dplyr::select(dplyr::all_of(target_name)) |>
  dplyr::pull() 
std_dev <- sd(raw_target)
mn <- mean(raw_target)

# unscale
y_obs <- (target_obs * std_dev) + mn
y_pred <- (target_pred *std_dev) + mn

plot(
  y_obs, y_pred, main = "Model evaluation", 
  xlab = "Observed target value", 
  ylab = "Predicted target value", 
  axes = FALSE, pch = 16
)
segments(
  x0=0, y0=0, x1=max(y_obs), y1=max(y_obs), lty = 2, lwd = 4, col = "red"
)
axis(1)
axis(2, las = 2)

# Pearson correlation
cor(y_obs, y_pred, method = "pearson")
# Spearman correlation
cor(y_obs, y_pred, method = "spearman")

#new example ####################################################
# Load the dataset
data <- as.matrix(mtcars[, -1])  # Use all features except for mpg
target <- mtcars$mpg  # Target is mpg

# Normalize the data (standard scaling)
data <- scale(data)
target <- scale(target)

# Initialize Parameters
set.seed(123)
input_size <- ncol(data)
hidden_size <- 10  # Increased hidden size for more complexity
output_size <- 1  # Single output for regression

# Initialize weights and biases
W1 <- matrix(rnorm(input_size * hidden_size, 0, 0.1), nrow = input_size, ncol = hidden_size)
b1 <- matrix(0, nrow = 1, ncol = hidden_size)

W2 <- matrix(rnorm(hidden_size * output_size, 0, 0.1), nrow = hidden_size, ncol = output_size)
b2 <- matrix(0, nrow = 1, ncol = output_size)

# Activation Function: Sigmoid
sigmoid <- function(x) {
  1 / (1 + exp(-x))
}

# Sigmoid Derivative
sigmoid_derivative <- function(x) {
  sigmoid(x) * (1 - sigmoid(x))
}

# Mean Squared Error Loss
mean_squared_error <- function(y_true, y_pred) {
  mean((y_true - y_pred)^2)
}

# Forward Pass
forward_pass <- function(x) {
  z1 <- x %*% W1 + matrix(rep(b1, nrow(x)), nrow = nrow(x), byrow = TRUE)
  a1 <- sigmoid(z1)
  y_pred <- a1 %*% W2 + matrix(rep(b2, nrow(a1)), nrow = nrow(a1), byrow = TRUE)
  
  return(list(y_pred = y_pred, a1 = a1, z1 = z1))
}

# Backpropagation
backpropagation <- function(x, y, forward_cache, learning_rate = 0.01) {
  y_pred <- forward_cache$y_pred
  a1 <- forward_cache$a1
  z1 <- forward_cache$z1
  
  # Loss Derivative
  dL_dy_pred <- 2 * (y_pred - y) / nrow(y)
  
  # Gradients for W2 and b2
  dL_dW2 <- t(a1) %*% dL_dy_pred
  dL_db2 <- colSums(dL_dy_pred)
  
  # Gradients for W1 and b1
  dL_da1 <- dL_dy_pred %*% t(W2)
  dL_dz1 <- dL_da1 * sigmoid_derivative(z1)
  
  dL_dW1 <- t(x) %*% dL_dz1
  dL_db1 <- colSums(dL_dz1)
  
  # Update parameters
  W1 <<- W1 - learning_rate * dL_dW1
  b1 <<- b1 - learning_rate * dL_db1
  W2 <<- W2 - learning_rate * dL_dW2
  b2 <<- b2 - learning_rate * dL_db2
}

# Training the Network
num_epochs <- 5000  # Increase number of epochs
learning_rate <- 0.01  # Adjust the learning rate if needed

for (i in 1:num_epochs) {
  forward_cache <- forward_pass(data)
  y_pred <- forward_cache$y_pred
  
  if (i %% 500 == 0) {
    loss <- mean_squared_error(target, y_pred)
    cat("Epoch:", i, "Loss:", loss, "\n")
  }
  
  backpropagation(data, target, forward_cache, learning_rate)
}

# Final Predictions
final_predictions <- forward_pass(data)$y_pred

# Reverse the standard scaling to interpret results
final_predictions <- final_predictions * attr(target, "scaled:scale") + attr(target, "scaled:center")

# Plot the Results
plot(mtcars$mpg, final_predictions, main = "Actual vs Predicted MPG",
     xlab = "Actual MPG", ylab = "Predicted MPG")
abline(0, 1, col = "red", lwd = 2)


#new example with packages
library(mlbench)
data("BreastCancer")

#Clean off rows with missing data
BreastCancer = BreastCancer[which(complete.cases(BreastCancer)==TRUE),]

head(BreastCancer)

names(BreastCancer)

y = as.matrix(BreastCancer[,11])
y[which(y=="benign")] = 0
y[which(y=="malignant")] = 1
y = as.numeric(y)
x = as.numeric(as.matrix(BreastCancer[,2:10]))
x = matrix(as.numeric(x),ncol=9)

library(deepnet)
nn <- nn.train(x, y, hidden = c(5))
yy = nn.predict(nn, x)
print(head(yy))

yhat = matrix(0,length(yy),1)
yhat[which(yy > mean(yy))] = 1
yhat[which(yy <= mean(yy))] = 0
cm = table(y,yhat)
print(cm)

print(sum(diag(cm))/sum(cm))

library(neuralnet)
df = data.frame(cbind(x,y))
nn = neuralnet(y~V1+V2+V3+V4+V5+V6+V7+V8+V9,data=df,hidden = 5)
yy = nn$net.result[[1]]
yhat = matrix(0,length(y),1)
yhat[which(yy > mean(yy))] = 1
yhat[which(yy <= mean(yy))] = 0
print(table(y,yhat))

plot(nn)

