# 11/7 CSC 401 Examples

# Load the iris dataset
data(iris)
df <- iris[, 1:2]  # Use only Sepal.Length and Sepal.Width for simplicity

# Step 1: Create Euclidean Distance Matrix
euclidean_distance <- function(data) {
  dist_matrix <- as.matrix(dist(data, method = "euclidean"))
  return(dist_matrix)
}

# Step 2: Perform Single Linkage Agglomerative Clustering
single_linkage_clustering <- function(dist_matrix) {
  n <- nrow(dist_matrix)
  clusters <- list()
  cluster_merges <- list()
  merge_heights <- c()
  
  for (i in 1:n) {
    clusters[[i]] <- i
  }
  
  while (length(clusters) > 1) {
    min_dist <- Inf
    closest_clusters <- c(NA, NA)
    for (i in 1:(length(clusters) - 1)) {
      for (j in (i + 1):length(clusters)) {
        cluster_i <- clusters[[i]]
        cluster_j <- clusters[[j]]
        distance <- min(dist_matrix[cluster_i, cluster_j])
        if (distance < min_dist) {
          min_dist <- distance
          closest_clusters <- c(i, j)
        }
      }
    }
    
    merge_heights <- c(merge_heights, min_dist)
    cluster_merges <- append(cluster_merges, list(closest_clusters))
    clusters[[closest_clusters[1]]] <- c(clusters[[closest_clusters[1]]], clusters[[closest_clusters[2]]])
    clusters[[closest_clusters[2]]] <- NULL
  }
  
  return(list(cluster_merges = cluster_merges, merge_heights = merge_heights, n = n))
}

# Step 3: Use Iris Data and Remove Setosa
data(iris)
iris_data <- iris[iris$Species != "setosa", -5]

# Step 4: Apply the Clustering
dist_matrix <- euclidean_distance(iris_data)
clustering_result <- single_linkage_clustering(dist_matrix)

# Step 5: Create a Dendrogram
plot_dendrogram <- function(clustering_result) {
  n <- clustering_result$n
  heights <- clustering_result$merge_heights
  merges <- clustering_result$cluster_merges
  
  # Initialize plot with adjusted height
  plot(0, type = "n", xlim = c(1, n), ylim = c(0, max(heights) * 1.1), 
       xlab = "Index", ylab = "Height", main = "Dendrogram")
  
  cluster_positions <- 1:n
  previous_heights <- rep(0, n)
  
  for (i in seq_along(merges)) {
    merge <- merges[[i]]
    left <- min(cluster_positions[merge])
    right <- max(cluster_positions[merge])
    mid <- mean(c(left, right))
    
    # Draw the horizontal line for the current merge height
    segments(left, heights[i], right, heights[i])
    
    # Draw the vertical lines to the merge height
    segments(left, previous_heights[left], left, heights[i])
    segments(right, previous_heights[right], right, heights[i])
    
    # Update positions and heights
    cluster_positions[merge] <- mid
    previous_heights[mid] <- heights[i]
  }
}

# Step 6: Plot the Dendrogram
plot_dendrogram(clustering_result)


#change distance and linkage
# Step 1: Create Manhattan Distance Matrix
manhattan_distance <- function(data) {
  dist_matrix <- as.matrix(dist(data, method = "manhattan"))
  return(dist_matrix)
}

# Step 2: Perform Average Linkage Agglomerative Clustering
average_linkage_clustering <- function(dist_matrix) {
  n <- nrow(dist_matrix)
  clusters <- list()
  cluster_merges <- list()
  merge_heights <- c()
  
  for (i in 1:n) {
    clusters[[i]] <- i
  }
  
  while (length(clusters) > 1) {
    min_avg_dist <- Inf
    closest_clusters <- c(NA, NA)
    for (i in 1:(length(clusters) - 1)) {
      for (j in (i + 1):length(clusters)) {
        cluster_i <- clusters[[i]]
        cluster_j <- clusters[[j]]
        
        # Calculate average linkage distance
        distance <- mean(dist_matrix[cluster_i, cluster_j])
        
        if (distance < min_avg_dist) {
          min_avg_dist <- distance
          closest_clusters <- c(i, j)
        }
      }
    }
    
    merge_heights <- c(merge_heights, min_avg_dist)
    cluster_merges <- append(cluster_merges, list(closest_clusters))
    clusters[[closest_clusters[1]]] <- c(clusters[[closest_clusters[1]]], clusters[[closest_clusters[2]]])
    clusters[[closest_clusters[2]]] <- NULL
  }
  
  return(list(cluster_merges = cluster_merges, merge_heights = merge_heights, n = n))
}

# Step 3: Use Iris Data and Remove Setosa
data(iris)
iris_data <- iris[iris$Species != "setosa", -5]

# Step 4: Apply the Clustering
dist_matrix <- manhattan_distance(iris_data)
clustering_result <- average_linkage_clustering(dist_matrix)

# Step 5: Create a Dendrogram
plot_dendrogram <- function(clustering_result) {
  n <- clustering_result$n
  heights <- clustering_result$merge_heights
  merges <- clustering_result$cluster_merges
  
  # Initialize plot with adjusted height
  plot(0, type = "n", xlim = c(1, n), ylim = c(0, max(heights) * 1.1), 
       xlab = "Index", ylab = "Height", main = "Dendrogram")
  
  cluster_positions <- 1:n
  previous_heights <- rep(0, n)
  
  for (i in seq_along(merges)) {
    merge <- merges[[i]]
    left <- min(cluster_positions[merge])
    right <- max(cluster_positions[merge])
    mid <- mean(c(left, right))
    
    # Draw the horizontal line for the current merge height
    segments(left, heights[i], right, heights[i])
    
    # Draw the vertical lines to the merge height
    segments(left, previous_heights[left], left, heights[i])
    segments(right, previous_heights[right], right, heights[i])
    
    # Update positions and heights
    cluster_positions[merge] <- mid
    previous_heights[mid] <- heights[i]
  }
}

# Step 6: Plot the Dendrogram
plot_dendrogram(clustering_result)

#linkage methods
complete_linkage <- function(cluster_i, cluster_j, dist_matrix) {
  max_dist <- max(dist_matrix[cluster_i, cluster_j])
  return(max_dist)
}

centroid_linkage <- function(cluster_i, cluster_j, data) {
  centroid_i <- colMeans(data[cluster_i, ])
  centroid_j <- colMeans(data[cluster_j, ])
  centroid_dist <- sqrt(sum((centroid_i - centroid_j)^2))
  return(centroid_dist)
}

wards_linkage <- function(cluster_i, cluster_j, data) {
  combined_cluster <- rbind(data[cluster_i, , drop = FALSE], data[cluster_j, , drop = FALSE])
  total_within_ssq <- sum((data[cluster_i, , drop = FALSE] - colMeans(data[cluster_i, , drop = FALSE]))^2) +
    sum((data[cluster_j, , drop = FALSE] - colMeans(data[cluster_j, , drop = FALSE]))^2)
  combined_within_ssq <- sum((combined_cluster - colMeans(combined_cluster))^2)
  increase_in_ssq <- combined_within_ssq - total_within_ssq
  return(increase_in_ssq)
}

median_linkage <- function(cluster_i, cluster_j, data) {
  median_i <- apply(data[cluster_i, , drop = FALSE], 2, median)
  median_j <- apply(data[cluster_j, , drop = FALSE], 2, median)
  median_dist <- sqrt(sum((median_i - median_j)^2))
  return(median_dist)
}

# Example linkage usage in clustering loop
for (i in 1:(length(clusters) - 1)) {
  for (j in (i + 1):length(clusters)) {
    cluster_i <- clusters[[i]]
    cluster_j <- clusters[[j]]
    
    # Use the desired linkage method here
    distance <- complete_linkage(cluster_i, cluster_j, dist_matrix)  # Replace with desired method
    
    if (distance < min_dist) {
      min_dist <- distance
      closest_clusters <- c(i, j)
    }
  }
}

