# Example for 9/10 CSC 401

# Objective Function: Mean Squared Error
mse <- function(y, y_pred) {
  mean((y - y_pred)^2)
}

# Gradient Descent Function
gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 1000) {
  m <- length(y)
  theta <- 0  # Start with an initial value of 0 for theta
  
  for (i in 1:n_iter) {
    y_pred <- theta * x
    gradient <- -(2/m) * sum(x * (y - y_pred))
    theta <- theta - learning_rate * gradient
  }
  
  return(theta)
}

# Example data
x <- 1:10
y <- 2 * x + 1  # Linear relationship with some noise

# Apply Gradient Descent
theta_optimal <- gradient_descent(x, y)
print(theta_optimal)

# Stochastic Gradient Descent Function
stochastic_gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 100) {
  m <- length(y)
  theta <- 0  # Start with an initial value of 0 for theta
  
  for (i in 1:n_iter) {
    for (j in 1:m) {
      y_pred <- theta * x[j]
      gradient <- -(2/m) * x[j] * (y[j] - y_pred)
      theta <- theta - learning_rate * gradient
    }
  }
  
  return(theta)
}

# Apply Stochastic Gradient Descent
theta_sgd <- stochastic_gradient_descent(x, y)
print(theta_sgd)

# Mini-Batch Gradient Descent Function
mini_batch_gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 100, batch_size = 3) {
  m <- length(y)
  theta <- 0  # Start with an initial value of 0 for theta
  
  for (i in 1:n_iter) {
    indices <- sample(1:m, batch_size)
    x_batch <- x[indices]
    y_batch <- y[indices]
    
    y_pred <- theta * x_batch
    gradient <- -(2/batch_size) * sum(x_batch * (y_batch - y_pred))
    theta <- theta - learning_rate * gradient
  }
  
  return(theta)
}

# Apply Mini-Batch Gradient Descent
theta_mini_batch <- mini_batch_gradient_descent(x, y)
print(theta_mini_batch)

# Newton's Method for a quadratic function
newtons_method <- function(x, y, n_iter = 10) {
  theta <- 0  # Initial guess
  for (i in 1:n_iter) {
    grad <- -2 * sum(x * (y - theta * x))
    hessian <- 2 * sum(x^2)
    theta <- theta - grad / hessian
  }
  return(theta)
}

# Apply Newton's Method
theta_newton <- newtons_method(x, y)
print(theta_newton)

# Adam Optimizer (simplified version)
adam_optimizer <- function(x, y, learning_rate = 0.01, beta1 = 0.9, beta2 = 0.999, epsilon = 1e-8, n_iter = 1000) {
  m <- length(y)
  theta <- 0
  m_t <- 0
  v_t <- 0
  for (t in 1:n_iter) {
    y_pred <- theta * x
    grad <- -(2/m) * sum(x * (y - y_pred))
    
    m_t <- beta1 * m_t + (1 - beta1) * grad
    v_t <- beta2 * v_t + (1 - beta2) * grad^2
    
    m_t_hat <- m_t / (1 - beta1^t)
    v_t_hat <- v_t / (1 - beta2^t)
    
    theta <- theta - learning_rate * m_t_hat / (sqrt(v_t_hat) + epsilon)
  }
  
  return(theta)
}

# Apply Adam Optimizer
theta_adam <- adam_optimizer(x, y)
print(theta_adam)

#custom gradient descent with decaying learning rate + check
gradient_descent_decay <- function(x, y, initial_lr = 0.1, decay = 0.001, n_iter = 1000) {
  m <- length(y)
  theta <- 0  # Start with an initial value of 0 for theta
  
  for (i in 1:n_iter) {
    y_pred <- theta * x
    gradient <- -(2/m) * sum(x * (y - y_pred))
    
    # Update learning rate
    learning_rate <- initial_lr / (1 + decay * i)
    
    # Update theta
    theta <- theta - learning_rate * gradient
    
    # Print theta to debug if it goes to NaN
    if (is.nan(theta)) {
      print(paste("Theta became NaN at iteration:", i))
      break
    }
  }
  
  return(theta)
}

# Example data
x <- 1:10
y <- 2 * x + 1  # Linear relationship

# Apply Custom Gradient Descent
theta_custom <- gradient_descent_decay(x, y, initial_lr = 0.01, decay = 0.0001)
print(theta_custom)

#more examples
data(mtcars)
mtcars_data <- mtcars[, c("mpg", "disp", "hp", "wt", "qsec", "drat")]  # Selecting 5 variables + mpg

# Normalizing the data for better convergence
normalize <- function(x) {
  return((x - mean(x)) / sd(x))
}

mtcars_data <- as.data.frame(lapply(mtcars_data, normalize))

# Separate predictors and target
x <- as.matrix(mtcars_data[, -1])
y <- mtcars_data$mpg

# Add a column of ones for the intercept
x <- cbind(1, x)

gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 1000) {
  m <- nrow(x)
  theta <- rep(0, ncol(x))  # Initialize theta
  
  for (i in 1:n_iter) {
    y_pred <- x %*% theta
    gradient <- -(2/m) * t(x) %*% (y - y_pred)
    theta <- theta - learning_rate * gradient
  }
  
  return(theta)
}

theta_gd <- gradient_descent(x, y)
print(theta_gd)

stochastic_gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 100) {
  m <- nrow(x)
  theta <- rep(0, ncol(x))  # Initialize theta
  
  for (i in 1:n_iter) {
    for (j in 1:m) {
      idx <- sample(1:m, 1)
      x_i <- x[idx, , drop = FALSE]
      y_i <- y[idx]
      
      y_pred <- x_i %*% theta
      gradient <- -(2) * t(x_i) %*% (y_i - y_pred)
      theta <- theta - learning_rate * gradient
    }
  }
  
  return(theta)
}

theta_sgd <- stochastic_gradient_descent(x, y)
print(theta_sgd)

mini_batch_gradient_descent <- function(x, y, learning_rate = 0.01, n_iter = 100, batch_size = 5) {
  m <- nrow(x)
  theta <- rep(0, ncol(x))  # Initialize theta
  
  for (i in 1:n_iter) {
    indices <- sample(1:m, batch_size)
    x_batch <- x[indices, , drop = FALSE]
    y_batch <- y[indices]
    
    y_pred <- x_batch %*% theta
    gradient <- -(2/batch_size) * t(x_batch) %*% (y_batch - y_pred)
    theta <- theta - learning_rate * gradient
  }
  
  return(theta)
}

theta_mbgd <- mini_batch_gradient_descent(x, y)
print(theta_mbgd)

newtons_method <- function(x, y, n_iter = 10) {
  m <- nrow(x)
  theta <- rep(0, ncol(x))  # Initialize theta
  
  for (i in 1:n_iter) {
    y_pred <- x %*% theta
    gradient <- -(2/m) * t(x) %*% (y - y_pred)
    hessian <- (2/m) * t(x) %*% x
    
    theta <- theta - solve(hessian) %*% gradient
  }
  
  return(theta)
}

theta_newton <- newtons_method(x, y)
print(theta_newton)

adam_optimizer <- function(x, y, learning_rate = 0.01, beta1 = 0.9, beta2 = 0.999, epsilon = 1e-8, n_iter = 1000) {
  m <- nrow(x)
  theta <- rep(0, ncol(x))  # Initialize theta
  m_t <- rep(0, ncol(x))
  v_t <- rep(0, ncol(x))
  
  for (t in 1:n_iter) {
    y_pred <- x %*% theta
    gradient <- -(2/m) * t(x) %*% (y - y_pred)
    
    m_t <- beta1 * m_t + (1 - beta1) * gradient
    v_t <- beta2 * v_t + (1 - beta2) * (gradient^2)
    
    m_t_hat <- m_t / (1 - beta1^t)
    v_t_hat <- v_t / (1 - beta2^t)
    
    theta <- theta - learning_rate * m_t_hat / (sqrt(v_t_hat) + epsilon)
  }
  
  return(theta)
}

theta_adam <- adam_optimizer(x, y)
print(theta_adam)

