# example from 9/12 CSC 401

# Create a matrix with numbers from 1 to 6
A <- matrix(1:6, nrow = 2, ncol = 3)
print(A)

# Creating a 3x3 matrix
B <- matrix(c(1, 2, 3, 4, 5, 6, 7, 8, 9), nrow = 3, ncol = 3)
print(B)

# Creating two matrices
A <- matrix(c(1, 2, 3, 4), nrow = 2, ncol = 2)
B <- matrix(c(5, 6, 7, 8), nrow = 2, ncol = 2)

# Addition
C <- A + B
print(C)

# Subtraction
D <- A - B
print(D)

# Matrix multiplication (element-wise)
E <- A * B
print(E)

# Matrix multiplication (true matrix multiplication)
F <- A %*% B
print(F)

# Transposing a matrix
G <- t(A)
print(G)

install.packages("Matrix")  # Install the package if you haven't already
library(Matrix)

# Create a non-singular matrix
B <- matrix(c(2, 1, 1, 
              4, -6, 0, 
              -2, 7, 2), nrow = 3, byrow = TRUE)

# Perform LU Decomposition
LU_decomp <- lu(B)

# Extract the L (lower triangular) and U (upper triangular) matrices
L <- expand(lu(B))$L
U <- expand(lu(B))$U
P <- expand(lu(B))$P  # The permutation matrix

# Display the matrices
print("Matrix L:")
print(L)

print("Matrix U:")
print(U)

print("Permutation Matrix P:")
print(P)

# Reconstructing the original matrix B
B_reconstructed <- P %*% L %*% U
print("Reconstructed Matrix B:")
print(B_reconstructed)

# QR Decomposition
QR_decomp <- qr(B)
Q <- qr.Q(QR_decomp)
R <- qr.R(QR_decomp)

print(Q)
print(R)

# Create a 4x3 matrix
A <- matrix(c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12), nrow = 4, byrow = TRUE)
print("Matrix A:")
print(A)

# Perform Singular Value Decomposition
svd_A <- svd(A)

# U matrix
U <- svd_A$u
print("Matrix U:")
print(U)

# Sigma (Diagonal) matrix
Sigma <- diag(svd_A$d)
print("Matrix Sigma:")
print(Sigma)

# V^T matrix
V_T <- svd_A$v
print("Matrix V^T:")
print(t(V_T))

# Reconstruct A using U, Sigma, and V^T
A_reconstructed <- U %*% Sigma %*% t(V_T)
print("Reconstructed Matrix A:")
print(A_reconstructed)

# Using the 'volcano' dataset for image compression example
image_matrix <- as.matrix(volcano)

# Perform SVD
svd_image <- svd(image_matrix)

# Retain only the top 20 singular values for compression
k <- 20
U_k <- svd_image$u[, 1:k]
Sigma_k <- diag(svd_image$d[1:k])
V_k_T <- svd_image$v[, 1:k]

# Reconstruct the compressed image
compressed_image <- U_k %*% Sigma_k %*% t(V_k_T)

# Plot the original and compressed images
par(mfrow = c(1, 2))
image(volcano, main = "Original Image")
image(compressed_image, main = "Compressed Image")
par(mfrow = c(1, 1))

# Simulate some data
set.seed(987)
x <- matrix(rnorm(100), nrow = 50, ncol = 2)
y <- 1 + x[,1] * 2 + x[,2] * 3 + rnorm(50, 0, 0.1)

# Add a column of 1s for the intercept
X <- cbind(1, x)

# Compute theta using the Normal Equation
theta <- solve(t(X) %*% X) %*% t(X) %*% y
print(theta)

