# 9/17 CSC 401 examples

# Load mtcars dataset
data(mtcars)

# Extract variables
X <- as.matrix(cbind(1, mtcars$hp))  # Add a column of ones for the intercept
y <- mtcars$mpg

# Normal Equation: beta = (X^T * X)^(-1) * X^T * y
beta_hat <- solve(t(X) %*% X) %*% t(X) %*% y

# Print coefficients
print("Coefficients:")
print(beta_hat)

# Plot the data and the regression line
plot(mtcars$hp, mtcars$mpg, main = "Simple Linear Regression: mpg ~ hp", xlab = "Horsepower", ylab = "Miles Per Gallon")
abline(beta_hat[1], beta_hat[2], col = "blue")

model_lm <- lm(mpg ~ hp, data = mtcars)
summary(model_lm)

# Extract variables for multiple regression
X_multi <- as.matrix(cbind(1, mtcars$hp, mtcars$wt, mtcars$qsec))
y_multi <- mtcars$mpg

# Normal Equation for Multiple Regression
beta_hat_multi <- solve(t(X_multi) %*% X_multi) %*% t(X_multi) %*% y_multi

# Print coefficients
print("Coefficients for Multiple Regression:")
print(beta_hat_multi)

y_pred <- X_multi %*% beta_hat_multi

plot(y_multi, y_pred, main = "Predicted vs Actual MPG", xlab = "Actual MPG", ylab = "Predicted MPG")
abline(0, 1, col = "red")  # Line of perfect prediction

model_lm_multi <- lm(mpg ~ hp + wt + qsec, data = mtcars)
summary(model_lm_multi)

# Setup for quadratic regression
X_poly <- as.matrix(cbind(1, mtcars$hp, mtcars$hp^2))
y_poly <- mtcars$mpg

# Solve using the normal equation
beta_hat_poly <- solve(t(X_poly) %*% X_poly) %*% t(X_poly) %*% y_poly

# Print coefficients
print("Coefficients for Polynomial Regression:")
print(beta_hat_poly)

# Plot the data and the regression curve
plot(mtcars$hp, mtcars$mpg, main = "Polynomial Regression: mpg ~ hp + hp^2", xlab = "Horsepower", ylab = "Miles Per Gallon")
curve(beta_hat_poly[1] + beta_hat_poly[2] * x + beta_hat_poly[3] * x^2, add = TRUE, col = "green")

model_lm_poly <- lm(mpg ~ hp + I(hp^2), data = mtcars)
summary(model_lm_poly)

# Covariance matrix for the coefficients in the simple linear regression
sigma_squared <- sum((y - X %*% beta_hat)^2) / (nrow(X) - ncol(X))
cov_matrix <- sigma_squared * solve(t(X) %*% X)

print("Covariance Matrix for Simple Linear Regression:")
print(cov_matrix)
