# 9/19 examples CSC 401

# Load the mtcars dataset
data(mtcars)

# Simple Linear Regression: mpg ~ hp
model_lm <- lm(mpg ~ hp, data = mtcars)

# Multiple Linear Regression: mpg ~ hp + wt + qsec
model_lm_multi <- lm(mpg ~ hp + wt + qsec, data = mtcars)

# Generalized Linear Model: mpg ~ hp + wt + qsec (using Gaussian family)
model_glm <- glm(mpg ~ hp + wt + qsec, family = gaussian, data = mtcars)

# LOESS (Locally Estimated Scatterplot Smoothing): mpg ~ hp
model_loess <- loess(mpg ~ hp, data = mtcars)

# Penalized Regression: Ridge Regression (L2) (using glmnet package)
# install.packages("glmnet")
library(glmnet)
x <- as.matrix(mtcars[, c("hp", "wt", "qsec")])
y <- mtcars$mpg
model_ridge <- glmnet(x, y, alpha = 0, lambda = 0.1) # Alpha=0 for Ridge, lambda controls regularization

# Spline Regression: mpg ~ hp (using splines)
library(splines)
model_spline <- lm(mpg ~ ns(hp, df = 4), data = mtcars)

# Gaussian Process Regression (using kernlab package)
# install.packages("kernlab")
library(kernlab)
model_gp <- gausspr(mpg ~ hp + wt + qsec, data = mtcars, kernel = "rbfdot")

# Get predicted values from the model
predicted_lm <- predict(model_lm)
predicted_multi <- predict(model_lm_multi)

# Calculate residuals by hand: Residual = Observed - Predicted
residuals_lm <- mtcars$mpg - predicted_lm
residuals_multi <- mtcars$mpg - predicted_multi

# Calculate Residual Sum of Squares (RSS)
RSS_lm <- sum(residuals_lm^2)

# Calculate Total Sum of Squares (TSS)
TSS <- sum((mtcars$mpg - mean(mtcars$mpg))^2)

# Calculate R-squared
R_squared_lm <- 1 - RSS_lm / TSS
R_squared_lm

# Number of observations
n <- length(mtcars$mpg)

# Calculate RMSE
RMSE_lm <- sqrt(RSS_lm / n)
RMSE_lm

# Calculate MAPE
MAPE_lm <- mean(abs((mtcars$mpg - predicted_lm) / mtcars$mpg)) * 100
MAPE_lm

# Number of predictors in the model
p_lm <- length(coefficients(model_lm))

# Calculate AIC
AIC_lm <- n * log(RSS_lm / n) + 2 * p_lm
AIC_lm

# Calculate BIC
BIC_lm <- n * log(RSS_lm / n) + p_lm * log(n)
BIC_lm

# Calculate error variance (sigma^2) from the full model (multiple regression)
sigma2_full <- sum(residuals_multi^2) / (n - p_lm)

# Calculate Mallow's Cp
Cp <- RSS_lm / sigma2_full + 2 * p_lm - n
Cp

# R-squared
summary(model_lm)$r.squared

# RMSE
sqrt(mean(residuals_lm^2))

# MAPE (not built-in, but you can verify)
mean(abs((mtcars$mpg - predicted_lm) / mtcars$mpg)) * 100

# AIC
AIC(model_lm)

# BIC
BIC(model_lm)

# Load the dataset
data(mtcars)

# Fit a simple linear regression model
model <- lm(mpg ~ wt + hp, data = mtcars)

# Calculate residuals
residuals <- resid(model)

# Plot Residuals vs Fitted Values
plot(fitted(model), residuals,
     main = "Residuals vs Fitted Values",
     xlab = "Fitted Values",
     ylab = "Residuals")
abline(h = 0, col = "red")

# QQ Plot of Residuals
qqnorm(residuals, main = "QQ Plot of Residuals")
qqline(residuals, col = "red")

# Calculate Cook's Distance
cooks_distance <- cooks.distance(model)

# Plot Cook's Distance
plot(cooks_distance, type = "h", main = "Cook's Distance",
     ylab = "Cook's Distance")
abline(h = 4 / nrow(mtcars), col = "red") # Common threshold line

# Calculate Leverage (Hat Values)
leverage <- hatvalues(model)

# Plot Leverage
plot(leverage, type = "h", main = "Leverage (Hat Values)",
     ylab = "Leverage")
abline(h = 2 * mean(leverage), col = "red") # Common threshold

# Calculate Standardized Residuals
standardized_residuals <- rstandard(model)

# Calculate Studentized Residuals
studentized_residuals <- rstudent(model)

# Plot Standardized Residuals
plot(standardized_residuals, type = "h", main = "Standardized Residuals",
     ylab = "Standardized Residuals")
abline(h = c(-2, 2), col = "red") # Common threshold lines

# Plot Studentized Residuals
plot(studentized_residuals, type = "h", main = "Studentized Residuals",
     ylab = "Studentized Residuals")
abline(h = c(-2, 2), col = "red") # Common threshold lines

# Load the car package
library(car)

# Perform the Bonferroni Outlier Test
outlier_test <- outlierTest(model)

# Display results
print(outlier_test)

# Perform Shapiro-Wilk Test for Normality
shapiro_test <- shapiro.test(residuals)

# Display results
print(shapiro_test)

# Install the EnvStats package if you haven't already
install.packages("EnvStats")

# Load the package
library(EnvStats)

# Example data
data <- c(10.1, 10.5, 10.2, 10.6, 10.8, 11.2, 10.4, 10.7, 11.4, 15.2, 16.8, 17.3)

# Perform Rosner's test
# The `k` parameter specifies the maximum number of potential outliers to test for.
rosner_test_result <- rosnerTest(data, k = 3)

# Display the results
print(rosner_test_result)

# Perform Rosner's test on the 'mpg' column of the mtcars dataset
rosner_test_mtcars <- rosnerTest(mtcars$mpg, k = 3)

# Display the results
print(rosner_test_mtcars)

# Load the mtcars dataset
data(mtcars)

# Create a boxplot for the mpg variable
boxplot(mtcars$mpg,
        main = "Boxplot of Miles Per Gallon (mpg)",
        ylab = "Miles Per Gallon",
        col = "lightblue")

# Create a boxplot with labels for outliers
boxplot(mtcars$mpg,
        main = "Boxplot of Miles Per Gallon (mpg)",
        ylab = "Miles Per Gallon",
        col = "lightblue",
        outline = TRUE)

# Identify and label the outliers
outliers <- boxplot.stats(mtcars$mpg)$out
outlier_indices <- which(mtcars$mpg %in% outliers)

# Add text labels for outliers
text(x = rep(1, length(outliers)), y = outliers, 
     labels = rownames(mtcars)[outlier_indices], 
     pos = 4, cex = 0.8, col = "red")

# Extract outliers using boxplot.stats
outliers <- boxplot.stats(mtcars$mpg)$out
print(outliers)

# Create a boxplot for the residuals
boxplot(residuals,
        main = "Boxplot of Residuals",
        ylab = "Residuals",
        col = "lightblue")

# Create a boxplot with labels for outliers
boxplot(residuals,
        main = "Residuals",
        ylab = "Residuals",
        col = "lightblue",
        outline = TRUE)

# Identify and label the outliers
outliers <- boxplot.stats(residuals)$out
outlier_indices <- which(residuals %in% outliers)

# Add text labels for outliers
text(x = rep(1, length(outliers)), y = outliers, 
     labels = rownames(mtcars)[outlier_indices], 
     pos = 4, cex = 0.8, col = "red")

# Extract outliers using boxplot.stats
outliers <- boxplot.stats(residuals)$out
print(outliers)

# Load the mtcars dataset
data(mtcars)

# Remove row 20
mtcars <- mtcars[-20, ]

# Display the modified dataset to verify
print(mtcars)
