# 9/24 CSC 401 Examples

ridge_regression <- function(X, y, lambda) {
  # Adding intercept term
  X <- cbind(1, X)
  
  # Identity matrix
  I <- diag(ncol(X))
  I[1, 1] <- 0  # Don't penalize the intercept
  
  # Normal equation with ridge penalty
  beta_hat <- solve(t(X) %*% X + lambda * I) %*% t(X) %*% y
  return(beta_hat)
}

# Example usage with mtcars dataset
X <- as.matrix(mtcars[, c("wt", "hp", "qsec")])
y <- mtcars$mpg
lambda <- 1

ridge_coefficients <- ridge_regression(X, y, lambda)
print(ridge_coefficients)

lasso_regression <- function(X, y, lambda, max_iter = 1000, tol = 1e-4) {
  n <- nrow(X)
  p <- ncol(X)
  beta <- rep(0, p)
  for (iter in 1:max_iter) {
    beta_old <- beta
    for (j in 1:p) {
      X_j <- X[, j]
      residual <- y - X %*% beta + beta[j] * X_j
      rho <- sum(X_j * residual)
      if (rho < -lambda / 2) {
        beta[j] <- (rho + lambda / 2) / sum(X_j^2)
      } else if (rho > lambda / 2) {
        beta[j] <- (rho - lambda / 2) / sum(X_j^2)
      } else {
        beta[j] <- 0
      }
    }
    if (sum(abs(beta - beta_old)) < tol) break
  }
  return(beta)
}

# Example usage with mtcars dataset
X <- as.matrix(mtcars[, c("wt", "hp", "qsec")])
X <- scale(X)  # Standardize predictors
y <- scale(mtcars$mpg)
lambda <- 0.1

lasso_coefficients <- lasso_regression(X, y, lambda)
print(lasso_coefficients)

elastic_net <- function(X, y, alpha, lambda, max_iter = 1000, tol = 1e-4) {
  # Decompose the elastic net into a combination of ridge and lasso
  ridge_part <- (1 - alpha) * ridge_regression(X, y, lambda)
  lasso_part <- alpha * lasso_regression(X, y, lambda, max_iter, tol)
  
  # Combine the results
  beta_hat <- ridge_part + lasso_part
  return(beta_hat)
}

# Example usage with mtcars dataset
alpha <- 0.5  # 50% ridge, 50% lasso
elastic_net_coefficients <- elastic_net(X, y, alpha, lambda)
print(elastic_net_coefficients)

# Example: Standardizing predictors
X <- scale(mtcars[, c("wt", "hp", "qsec")])

# Load required libraries
library(splines)

# Create a function for penalized spline regression
penalized_spline <- function(x, y, lambda) {
  # Create B-spline basis
  spline_basis <- bs(x, degree = 3, df = 10) # cubic spline with 10 degrees of freedom
  
  # Set up the penalty matrix (penalize the second derivative)
  penalty_matrix <- diff(diag(ncol(spline_basis)), differences = 2)
  
  # Estimate the coefficients using penalized least squares
  beta <- solve(t(spline_basis) %*% spline_basis + lambda * t(penalty_matrix) %*% penalty_matrix) %*% t(spline_basis) %*% y
  
  # Return the fitted values
  fitted_values <- spline_basis %*% beta
  return(list(fitted_values = fitted_values, beta = beta))
}

# Apply the penalized spline to the mtcars dataset
x <- mtcars$wt
y <- mtcars$mpg
lambda <- 1

result <- penalized_spline(x, y, lambda)

# Plot the results
plot(x, y, main = "Penalized Spline Regression", xlab = "Weight (wt)", ylab = "Miles Per Gallon (mpg)")
lines(sort(x), result$fitted_values[order(x)], col = "red", lwd = 2)

lambda <- 10

result <- penalized_spline(x, y, lambda)

# Plot the results
plot(x, y, main = "Penalized Spline Regression", xlab = "Weight (wt)", ylab = "Miles Per Gallon (mpg)")
lines(sort(x), result$fitted_values[order(x)], col = "red", lwd = 2)

lambda <- 0.1

result <- penalized_spline(x, y, lambda)

# Plot the results
plot(x, y, main = "Penalized Spline Regression", xlab = "Weight (wt)", ylab = "Miles Per Gallon (mpg)")
lines(sort(x), result$fitted_values[order(x)], col = "red", lwd = 2)

# Load required library
library(splines)

# Generate B-spline basis functions for two variables
spline_basis_wt <- bs(mtcars$wt, degree = 3, df = 5)  # Spline for weight
spline_basis_hp <- bs(mtcars$hp, degree = 3, df = 5)  # Spline for horsepower

# Combine the spline basis functions into a design matrix
design_matrix <- cbind(spline_basis_wt, spline_basis_hp)

# Fit a linear model using the design matrix
model <- lm(mtcars$mpg ~ design_matrix)

# Summarize the model
summary(model)

# Generate predictions and plot
predicted_mpg <- predict(model)

# Plot the original data and fitted values
plot(mtcars$wt, mtcars$mpg, pch = 19, col = 'blue', xlab = 'Weight (wt)', ylab = 'Miles Per Gallon (mpg)')
points(mtcars$wt, predicted_mpg, col = 'red', pch = 19)

# Load mgcv library for generalized additive models (GAMs)
library(mgcv)

# Fit a GAM with tensor product splines for wt and hp
gam_model <- gam(mpg ~ te(wt, hp), data = mtcars)

# Summarize the model
summary(gam_model)

# Plot the smooth interaction between wt and hp
plot(gam_model, scheme = 1)

model <- smooth.spline(mtcars$wt, mtcars$mpg)
plot(mtcars$wt, mtcars$mpg, main="Smoothing Spline Example")
lines(model, col="red")


library(splines)
model <- lm(mpg ~ ns(wt, df=4), data=mtcars)
plot(mtcars$wt, mtcars$mpg)
lines(mtcars$wt, predict(model), col="red")


