#401 lecture examples 9/3

my_mean <- function(x) {
  sum(x) / length(x)
}

# Example usage
data <- c(1, 2, 3, 4, 5)
my_mean(data)

data<-mtcars$mpg
my_mean(data)

data<-mtcars$am
my_mean(data)

my_median <- function(x) {
  sorted_x <- sort(x)
  n <- length(x)
  if (n %% 2 == 1) {
    sorted_x[(n + 1) / 2]
  } else {
    (sorted_x[n / 2] + sorted_x[n / 2 + 1]) / 2
  }
}

# Example usage
my_median(data)

my_sd <- function(x) {
  m <- my_mean(x)
  sqrt(sum((x - m)^2) / (length(x) - 1))
}

# Example usage
my_sd(data)

my_table <- function(x) {
  unique_vals <- unique(x)
  freq <- sapply(unique_vals, function(val) sum(x == val))
  names(freq) <- unique_vals
  return(freq)
}

# Example usage
categorical_data <- c("A", "B", "A", "C", "B", "A")
my_table(categorical_data)

data<-mtcars$cyl
my_table(data)
table(mtcars$cyl)

my_aggregate_mean <- function(data, group) {
  unique_groups <- unique(group)
  group_means <- sapply(unique_groups, function(g) my_mean(data[group == g]))
  names(group_means) <- unique_groups
  return(group_means)
}

# Example usage
values <- c(5, 10, 15, 20, 25, 30)
groups <- c("A", "A", "B", "B", "C", "C")
my_aggregate_mean(values, groups)

aggregate(mpg ~ cyl, data = mtcars, FUN = my_mean)

mtcars[mtcars$cyl == 6, ]
mtcars[1:5, ]

my_t_test <- function(x, y) {
  mean_diff <- my_mean(x) - my_mean(y)
  pooled_sd <- sqrt(((length(x) - 1) * my_sd(x)^2 + (length(y) - 1) * my_sd(y)^2) / (length(x) + length(y) - 2))
  t_stat <- mean_diff / (pooled_sd * sqrt(1 / length(x) + 1 / length(y)))
  return(t_stat)
}
#this assumes the difference is assumed to be zero.

# Example usage
group1 <- c(5, 6, 7, 8, 9)
group2 <- c(1, 2, 3, 4, 5)
my_t_test(group1, group2)

my_correlation <- function(x, y) {
  cov_xy <- sum((x - my_mean(x)) * (y - my_mean(y))) / (length(x) - 1)
  cor_xy <- cov_xy / (my_sd(x) * my_sd(y))
  return(cor_xy)
}

# Example usage
var1 <- c(1, 2, 3, 4, 5)
var2 <- c(5, 4, 3, 2, 1)
my_correlation(var1, var2)

cor(mtcars$hp, mtcars$mpg)

#means
my_mean <- function(x) {
  sum(x) / length(x)
}

apply_means <- function(data) {
  sapply(data, my_mean)
}

# Example usage
data("mtcars")
apply_means(mtcars)

#median
my_median <- function(x) {
  sorted_x <- sort(x)
  n <- length(x)
  if (n %% 2 == 1) {
    sorted_x[(n + 1) / 2]
  } else {
    (sorted_x[n / 2] + sorted_x[n / 2 + 1]) / 2
  }
}

apply_medians <- function(data) {
  sapply(data, my_median)
}

# Example usage
apply_medians(mtcars)

#standard deviation (sample)
my_sd <- function(x) {
  m <- my_mean(x)
  sqrt(sum((x - m)^2) / (length(x) - 1))
}

apply_sds <- function(data) {
  sapply(data, my_sd)
}

# Example usage
apply_sds(mtcars)

#Summary table
my_table <- function(x) {
  unique_vals <- unique(x)
  freq <- sapply(unique_vals, function(val) sum(x == val))
  names(freq) <- unique_vals
  return(freq)
}

apply_tables <- function(data) {
  lapply(data, function(column) {
    if (is.factor(column) || is.character(column)) {
      my_table(column)
    } else {
      NULL
    }
  })
}

# Example usage
apply_tables(mtcars)

#Aggregate
my_aggregate_mean <- function(data, group) {
  unique_groups <- unique(group)
  group_means <- sapply(unique_groups, function(g) my_mean(data[group == g]))
  names(group_means) <- unique_groups
  return(group_means)
}

apply_aggregate_means <- function(data, group_column) {
  data <- data.frame(data)
  groups <- data[[group_column]]
  data <- data[ , !(names(data) %in% group_column)]
  lapply(data, function(column) my_aggregate_mean(column, groups))
}

# Example usage
apply_aggregate_means(mtcars, "cyl")

#T-tests
# Load data
data("mtcars")

# Define custom mean function
my_mean <- function(x) {
  sum(x) / length(x)
}

# Define custom variance function
my_variance <- function(x) {
  m <- my_mean(x)
  sum((x - m)^2) / (length(x) - 1)
}

# Define custom t-test function
my_t_test <- function(x, y) {
  mean_diff <- my_mean(x) - my_mean(y)
  pooled_sd <- sqrt(((length(x) - 1) * my_variance(x) + (length(y) - 1) * my_variance(y)) / (length(x) + length(y) - 2))
  t_stat <- mean_diff / (pooled_sd * sqrt(1 / length(x) + 1 / length(y)))
  return(t_stat)
}

# Function to apply t-test to each column in the data frame
apply_t_tests <- function(data, group_column) {
  data <- data.frame(data)
  groups <- data[[group_column]]
  data <- data[ , !(names(data) %in% group_column)]
  unique_groups <- unique(groups)
  
  if (length(unique_groups) != 2) {
    stop("The group column must have exactly 2 unique values.")
  }
  
  group1 <- data[groups == unique_groups[1], ]
  group2 <- data[groups == unique_groups[2], ]
  
  sapply(names(data), function(column) my_t_test(group1[[column]], group2[[column]]))
}

# Subset mtcars to include only cars with 4 and 6 cylinders
subset_mtcars <- mtcars[mtcars$cyl %in% c(4, 6), ]

# Example usage
apply_t_tests(subset_mtcars, "cyl")



my_correlation <- function(x, y) {
  cov_xy <- sum((x - my_mean(x)) * (y - my_mean(y))) / (length(x) - 1)
  cor_xy <- cov_xy / (my_sd(x) * my_sd(y))
  return(cor_xy)
}

#Correlations
apply_correlations <- function(data) {
  n <- ncol(data)
  cor_matrix <- matrix(NA, n, n)
  colnames(cor_matrix) <- rownames(cor_matrix) <- names(data)
  for (i in 1:(n - 1)) {
    for (j in (i + 1):n) {
      cor_matrix[i, j] <- cor_matrix[j, i] <- my_correlation(data[[i]], data[[j]])
    }
  }
  return(cor_matrix)
}

# Example usage
apply_correlations(mtcars)

my_skewness <- function(x) {
  n <- length(x)
  mean_x <- mean(x)
  sd_x <- sd(x)
  skewness <- (n/((n-1)*(n-2))) * sum(((x - mean_x) / sd_x)^3)
  return(skewness)
}

# Example usage
data <- c(1, 2, 3, 4, 5, 6, 7, 8, 9)
my_skewness(data)

my_kurtosis <- function(x) {
  n <- length(x)
  mean_x <- mean(x)
  sd_x <- sd(x)
  kurtosis <- (n*(n+1))/((n-1)*(n-2)*(n-3)) * sum(((x - mean_x) / sd_x)^4) - (3*(n-1)^2)/((n-2)*(n-3))
  return(kurtosis)
}

# Example usage
my_kurtosis(data)

my_cv <- function(x) {
  mean_x <- mean(x)
  sd_x <- sd(x)
  cv <- sd_x / mean_x
  return(cv)
}

# Example usage
my_cv(data)

my_mad <- function(x) {
  mean_x <- mean(x)
  mad <- mean(abs(x - mean_x))
  return(mad)
}

# Example usage
my_mad(data)

my_harmonic_mean <- function(x) {
  harmonic_mean <- length(x) / sum(1 / x)
  return(harmonic_mean)
}

# Example usage
my_harmonic_mean(data)

my_geometric_mean <- function(x) {
  geometric_mean <- prod(x)^(1/length(x))
  return(geometric_mean)
}

# Example usage
my_geometric_mean(data)

my_median_absolute_deviation <- function(x) {
  med_x <- median(x)
  mad <- median(abs(x - med_x))
  return(mad)
}

# Example usage
my_median_absolute_deviation(data)

my_trimmed_mean <- function(x, trim = 0.1) {
  sorted_x <- sort(x)
  n <- length(x)
  k <- floor(trim * n)
  trimmed_x <- sorted_x[(k + 1):(n - k)]
  trimmed_mean <- mean(trimmed_x)
  return(trimmed_mean)
}

# Example usage
my_trimmed_mean(data, 0.1)


