#9/5 CSC 401 Examples

euclidean_distance <- function(x, y) {
  sqrt(sum((x - y)^2))
}

# Select the first two rows (cars) and all columns except the 'cyl' column
car1 <- mtcars[1, -2]
car2 <- mtcars[2, -2]

# Calculate Euclidean distance
euclidean_distance(car1, car2)

manhattan_distance <- function(x, y) {
  sum(abs(x - y))
}

# Calculate Manhattan distance
manhattan_distance(car1, car2)

minkowski_distance <- function(x, y, p) {
  sum(abs(x - y)^p)^(1/p)
}

# Calculate Minkowski distance with p = 3
minkowski_distance(car1, car2, p = 3)


chebyshev_distance <- function(x, y) {
  max(abs(x - y))
}

# Calculate Chebyshev distance
chebyshev_distance(car1, car2)

cosine_similarity <- function(x, y) {
  sum(x * y) / (sqrt(sum(x^2)) * sqrt(sum(y^2)))
}

# Calculate Cosine Similarity
cosine_similarity(car1, car2)

jaccard_distance <- function(x, y) {
  intersection <- sum(x & y)
  union <- sum(x | y)
  1 - (intersection / union)
}

# Create binary vectors based on mpg (above/below threshold)
binary_car1 <- as.numeric(mtcars[1, ] > median(mtcars$mpg))
binary_car2 <- as.numeric(mtcars[2, ] > median(mtcars$mpg))

# Calculate Jaccard distance
jaccard_distance(binary_car1, binary_car2)

# Load the dataset
data("mtcars")

# Select the first two rows (cars) and all columns except the 'cyl' column
car1 <- as.numeric(mtcars[1, -2])
car2 <- as.numeric(mtcars[2, -2])

# Define Mahalanobis distance function
mahalanobis_distance <- function(x, y, cov_matrix) {
  diff <- as.matrix(x - y, ncol=1)  # Convert the difference to a column matrix
  sqrt(t(diff) %*% solve(cov_matrix) %*% diff)
}

# Calculate the covariance matrix of the entire dataset (excluding 'cyl')
cov_matrix <- cov(mtcars[ , -2])

# Calculate Mahalanobis distance
mahalanobis_distance(car1, car2, cov_matrix)

# Load the dataset
data("mtcars")

# Select only the numeric columns (excluding categorical or identifiers if necessary)
# Here we use the entire dataset without 'cyl' for example purposes.
mtcars_numeric <- mtcars[ , -2]

# Define Euclidean distance function
euclidean_distance <- function(x, y) {
  sqrt(sum((x - y)^2))
}

# Initialize an empty matrix to store the distances
num_rows <- nrow(mtcars_numeric)
distance_matrix <- matrix(NA, nrow = num_rows, ncol = num_rows)

# Calculate Euclidean distance for each pair of rows (cars)
for (i in 1:num_rows) {
  for (j in 1:num_rows) {
    distance_matrix[i, j] <- euclidean_distance(mtcars_numeric[i, ], mtcars_numeric[j, ])
  }
}

# Optionally add row and column names to the matrix for easier interpretation
rownames(distance_matrix) <- rownames(mtcars_numeric)
colnames(distance_matrix) <- rownames(mtcars_numeric)

# View the distance matrix
print(distance_matrix)

# Function to perform Min-Max Scaling
min_max_scaling <- function(x) {
  (x - min(x)) / (max(x) - min(x))
}

# Apply to mtcars dataset
mtcars_min_max <- as.data.frame(lapply(mtcars, min_max_scaling))

# View scaled data
head(mtcars_min_max)

# Function to perform Z-Score Scaling
z_score_scaling <- function(x) {
  (x - mean(x)) / sd(x)
}

# Apply to mtcars dataset
mtcars_z_scaled <- as.data.frame(lapply(mtcars, z_score_scaling))

# View scaled data
head(mtcars_z_scaled)

# Function to perform Max Abs Scaling
max_abs_scaling <- function(x) {
  x / max(abs(x))
}

# Apply to mtcars dataset
mtcars_max_abs <- as.data.frame(lapply(mtcars, max_abs_scaling))

# View scaled data
head(mtcars_max_abs)

# Function to perform Robust Scaling
robust_scaling <- function(x) {
  (x - median(x)) / IQR(x)
}

# Apply to mtcars dataset
mtcars_robust_scaled <- as.data.frame(lapply(mtcars, robust_scaling))

# View scaled data
head(mtcars_robust_scaled)

# Function to perform Decimal Scaling
decimal_scaling <- function(x) {
  j <- ceiling(log10(max(abs(x))))
  x / 10^j
}

# Apply to mtcars dataset
mtcars_decimal_scaled <- as.data.frame(lapply(mtcars, decimal_scaling))

# View scaled data
head(mtcars_decimal_scaled)


